unit RakBinaryStreamDataXML;

{///////////////////////////////////////////////////////////////////////////////
// Beschreibung : Die Unit RakBinaryStreamDataXML stellt Funktionen zur Verfgung,
                  mit denen das binre Format, das durch die Klassen der Unit
                  RakBinaryStreamData erzeugt werden kann, in ein XML-Format
                  konvertiert werden kann.
                  Damit diese Unit bersetzt werden kann, wird der XDOM von
                  Dieter Khler ("http://www.philo.de/homepage.htm") bentigt.
// erstellt am  : 14.07.2002
////////////////////////////////////////////////////////////////////////////////
// nderungen:
// nderungen:
    11.1.2003: An die Version 2.3.37 von XDOM angepasst
///////////////////////////////////////////////////////////////////////////////}
// Delphi 5 Implementation
//
// Copyright (c) 2001-2002 by Richard Kaspar
// ("http://www.kasparsoft.de")
//
// Definitions:
// - "Package" refers to the collection of files distributed by
//   the Copyright Holder, and derivatives of that collection of
//   files created through textual modification.
// - "Standard Version" refers to such a Package if it has not
//   been modified, or has been modified in accordance with the
//   wishes of the Copyright Holder.
// - "Copyright Holder" is whoever is name in the copyright or
//   copyrights for the package.
// - "You" is you, if you're thinking about copying or distributing
//   this Package.
//
// Permission is hereby granted, free of charge, to any person
// obtaining a copy of this software and associated documentation
// files (the "Package"), to deal in the Package without restriction,
// including without limitation the rights to use, copy, modify,
// merge, publish, distribute, sublicense, and/or sell copies of the
// Package, and to permit persons to whom the Package is furnished
// to do so, subject to the following conditions:
//
// The above copyright notice and this permission notice shall be
// included in all copies or substantial portions of the Package.
//
// You may modify your copy of this Package in any way, provided
// that you insert a prominent notice in each changed file stating
// how and when you changed a file, and provided that you do at
// least one of the following:
//
// a) allow the Copyright Holder to include your modifications in
// the Standard Version of the Package.
//
// b) use the modified Package only within your corporation or
// organization.
//
// c) rename any non standard executables, units, and classes so
// the names do not conflict with standard executables, units, and
// classes, and provide a separate manual page that clearly documents
// how it differs from the standard version.
//
// d) make other distribution arrangements with the Copyright Holder.
//
// The name of the Copyright Holder may not be used to endorse or
// promote products derived from this Package without specific prior
// written permission.
//
// THE PACKAGE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
// EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
// MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.
// IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY
// CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT,
// TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE
// PACKAGE OR THE USE OR OTHER DEALINGS IN THE PACKAGE.
{$I RAK.INC}


interface

uses RakBinaryStreamData, XDOM_2_3, miscfuncXML, miscfunc, classes, sysutils;

procedure XmlStreamToRakBinary(Xml: TStream; RakBinary: TRakBinaryStreamData);
procedure RakBinaryToXmlStream(RakBinary: TRakBinaryStreamData; Xml: TStream);

implementation

procedure XmlStreamToRakBinary(Xml: TStream; RakBinary: TRakBinaryStreamData);
  procedure ReadTag(BinaryTag: TRakBinaryTag; XmlTag: TdomElement);
  var
    Block: TdomElement;
    Node: TdomNode;
    Data: TdomElement;
    I: integer;
    BaseTypeId: integer;
    DataType: string;
    Attr: TRakBinaryAttribute;
    DataTypeClass: TRakBinaryDataType;
  begin
    BinaryTag.TagName := XmlTag.TagName;

    Block := FirstChildTag(XmlTag, 'Attributes');
    if Block <> nil then
    begin
      for I := 0 to Block.ChildNodes.Length - 1 do
      begin
        Node := Block.ChildNodes.Item(i);
        if (Node is TdomElement) and (TdomElement(Node).TagName = 'Attribute') then
        begin
          Data := TdomElement(Node);

          BaseTypeId := StrToInt(Data.getAttributeLiteralValue('BaseTypeId'));
          DataType := Data.getAttributeLiteralValue('DataType');
          if DataType = '' then
            Attr := BinaryTag.Attributes.Add(TRakBinaryAttribute.DataTypeOf(BaseTypeId)) as TRakBinaryAttribute
          else
          begin
            DataTypeClass := TRakBinaryAttribute.DataTypeOfClassName(DataType);
            if DataTypeClass = nil then
              Attr := BinaryTag.Attributes.Add(TRakBinaryAttribute.DataTypeOf(BaseTypeId)) as TRakBinaryAttribute
            else
              Attr := BinaryTag.Attributes.Add(DataTypeClass) as TRakBinaryAttribute;
          end;
          Attr.ItemName := Data.getAttributeLiteralValue('ItemName');

          Attr.IsNull := StrToBool(Data.getAttributeLiteralValue('IsNull'));
          if not Attr.IsNull then
          begin
            if Attr is TRakBinaryBlob then
              TRakBinaryBlob(Attr).AsBase64 := (Data.ChildNodes.Item(0) as TdomText).Data
            else
              Attr.AsWideString := (Data.ChildNodes.Item(0) as TdomText).Data;
          end;
        end;
      end;
    end;

    Block := FirstChildTag(XmlTag, 'Tags');
    if Block <> nil then
    begin
      for I := 0 to Block.ChildNodes.Length - 1 do
      begin
        Node := Block.ChildNodes.Item(i);

        if Node is TdomElement then
          ReadTag(BinaryTag.AddTag, TdomElement(Node));
      end;
    end;
  end;
var
  Puffer: string;
  doc: TdomDocument;
  Base: TdomElement;
  Data: TdomElement;
begin
  Xml.Position := 0;
  SetLength(Puffer, Xml.Size);
  Xml.Read(PChar(Puffer)^, Xml.Size);
  doc := CreateXmlDocument(Puffer);
  try
    Base := doc.documentElement;
    RakBinary.ApplicationName := Base.getAttributeLiteralValue('ApplicationName');
    RakBinary.ApplicationVersion := Base.getAttributeLiteralValue('ApplicationVersion');

    Data := FirstChildTag(Base, 'Data');
    if Data <> nil then
      ReadTag(RakBinary, Data);
  finally
    FreeXmlDocument(doc);
  end;
end;

procedure RakBinaryToXmlStream(RakBinary: TRakBinaryStreamData; Xml: TStream);
  procedure WriteTag(BinaryTag: TRakBinaryTag; XmlTag: TdomElement);
  var
    I: integer;
    SubTag: TRakBinaryTag;
    Attr: TRakBinaryAttribute;
    Block: TdomElement;
    Data: TdomElement;
    TextElem: TdomText;
  begin
    Block := XmlTag.OwnerDocument.CreateElement('Attributes');
    XmlTag.AppendChild(Block);
    for I := 0 to BinaryTag.Attributes.Count - 1 do
    begin
      Attr := BinaryTag.Attributes[I];
      Data := Block.OwnerDocument.CreateElement('Attribute');
      Block.AppendChild(Data);

      Data.SetAttribute('BaseTypeId', IntToStr(Attr.BaseTypeId));
      if Attr.ClassParent <> TRakBinaryAttribute then
      begin
        if TRakBinaryAttribute.DataTypeOfClassName(Attr.DataTypeAsString) = nil then
          raise Exception.Create('Die Klasse "' + Attr.DataTypeAsString + '" muss mit TRakBinaryAttribute.RegisterBinaryDataType registriert werden.');

        Data.SetAttribute('DataType', Attr.DataTypeAsString)
      end;

      Data.SetAttribute('ItemName', Attr.ItemName);
      Data.SetAttribute('IsNull', BoolToStr(Attr.IsNull));
      if Attr is TRakBinaryBlob then
        TextElem := Block.OwnerDocument.CreateTextNode(TRakBinaryBlob(Attr).AsBase64)
      else
        TextElem := Block.OwnerDocument.CreateTextNode(Attr.AsWideString);
      Data.AppendChild(TextElem);
    end;

    Block := XmlTag.OwnerDocument.CreateElement('Tags');
    XmlTag.AppendChild(Block);
    for I := 0 to BinaryTag.Count - 1 do
    begin
      SubTag := BinaryTag[I];
      Data := XmlTag.OwnerDocument.CreateElement(SubTag.TagName);
      Block.AppendChild(Data);
      WriteTag(SubTag, Data);
    end;
  end;
var
  doc: TdomDocument;
  Base: TdomElement;
  Data: TdomElement;
  Code: AnsiString;
  dxp: TDomToXmlParser;
begin
  doc := CreateXmlDocumentWithTag('RakBinaryStreamData');
  try
    Base := doc.documentElement;
    Base.SetAttribute('ApplicationName', RakBinary.ApplicationName);
    Base.SetAttribute('ApplicationVersion', RakBinary.ApplicationVersion);
    Data := Base.OwnerDocument.CreateElement('Data');
    Base.AppendChild(Data);
    WriteTag(RakBinary, Data);

    dxp := TDomToXmlParser.Create(nil);
    try
      dxp.domImpl := doc.domImplementation;
      dxp.writeToString(doc, 'Latin1', Code);
    finally
      dxp.Free;
    end;

  finally
    FreeXmlDocument(doc);
  end;
  Xml.Write(PChar(Code)^, Length(Code));
end;

end.
