unit RakConnectable;

{///////////////////////////////////////////////////////////////////////////////
// Beschreibung : Die Unit RakConnectable ist veraltet und nur aus
                  Kompalitts-Grnden vorhanden.
                  RakConnectable wird durch die Unit RakNetz ersetzt
// erstellt am  : 1.9.2001
////////////////////////////////////////////////////////////////////////////////
// nderungen:
///////////////////////////////////////////////////////////////////////////////}
// Delphi 5 Implementation
//
// Copyright (c) 2000 by Richard Kaspar
// ("http://www.kasparsoft.de")
//
// Definitions:
// - "Package" refers to the collection of files distributed by
//   the Copyright Holder, and derivatives of that collection of
//   files created through textual modification.
// - "Standard Version" refers to such a Package if it has not
//   been modified, or has been modified in accordance with the
//   wishes of the Copyright Holder.
// - "Copyright Holder" is whoever is name in the copyright or
//   copyrights for the package.
// - "You" is you, if you're thinking about copying or distributing
//   this Package.
//
// Permission is hereby granted, free of charge, to any person
// obtaining a copy of this software and associated documentation
// files (the "Package"), to deal in the Package without restriction,
// including without limitation the rights to use, copy, modify,
// merge, publish, distribute, sublicense, and/or sell copies of the
// Package, and to permit persons to whom the Package is furnished
// to do so, subject to the following conditions:
//
// The above copyright notice and this permission notice shall be
// included in all copies or substantial portions of the Package.
//
// You may modify your copy of this Package in any way, provided
// that you insert a prominent notice in each changed file stating
// how and when you changed a file, and provided that you do at
// least one of the following:
//
// a) allow the Copyright Holder to include your modifications in
// the Standard Version of the Package.
//
// b) use the modified Package only within your corporation or
// organization.
//
// c) rename any non standard executables, units, and classes so
// the names do not conflict with standard executables, units, and
// classes, and provide a separate manual page that clearly documents
// how it differs from the standard version.
//
// d) make other distribution arrangements with the Copyright Holder.
//
// The name of the Copyright Holder may not be used to endorse or
// promote products derived from this Package without specific prior
// written permission.
//
// THE PACKAGE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
// EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
// MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.
// IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY
// CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT,
// TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE
// PACKAGE OR THE USE OR OTHER DEALINGS IN THE PACKAGE.
{$I RAK.INC}

interface

uses Classes, sysutils,
  RakCollection, MultiEvent;

type
  TRakConnection = class;
  TRakConnectionCol = class;
  TRakConnectionColClass = class of TRakConnectionCol;
  TRakConnectableList = class;

  TRakConnectableStates = (
    rcsDestroying,     // Das Objekt zerstrt sich; Destroy wurde aufgerufen
    rcsDisconnecting,  // Das Objekt trennt sich von einem anderen Objekt
    rcsDisconnectingAll, // Das Objekt trennt sich von allen Objekten
    rcsConnecting,     // Das Objekt verbindet sich mit einem anderem Objekt
    rcsAutoDestroying, // Das Objekt zerstrt sich selbst
    rcsSearchingAnker, // Das Objekt sucht nach einem anderem Objekt
                       // mit dem es direkt oder indirekt verbunden ist
                       // und dessen Property Autodestroy gleich false ist
    rcsChangingNode1,  // Die Verbindung hngt ihr eines Ende an ein anderes Objekt
    rcsChangingNode2,  // Die Verbindung hngt ihr anderes Ende an ein anderes Objekt
    rcsWorking);       // Das Objekt macht irgendetwas; rcsWorking ist
                       // automatisch in State enthalten, sobald einer der
                       // anderen Stati in State aufgenommen wird.
                       // rcsWorking ist der einzige Status, der mitzhlt,
                       // wie oft er gesetzt und wieder gelscht wurde.
                       // Wenn man z.B. rcsWorking zweimal setzt und einmal
                       // lscht, dann steht immer noch rcsWorking in State.
                       // Bei den anderen Stati ist dies nicht so.
                       // Trotzdem mu man auch die anderen Stati so oft
                       // lschen wie man sie setzt, da beim Setzen eines der
                       // anderen Stati automatisch rcsWorking mit gesetzt wird
                       // und dann natrlich auch mit gelscht wird.
  TRakConnectableState = set of TRakConnectableStates;

  TRakConnectableClass = class of TRakConnectable;
  TRakConnectable = class
  private
    FAutoDestroy: boolean;
    FrcsWorking: integer;
    FState: TRakConnectableState;
    FConnections: TRakConnectionCol;

    function GetCount: integer;
    function GetConnections(Idx: integer): TRakConnection;
    function GetNodes(Idx: integer): TRakConnectable;
    function GetConnectionInfoObjs(Idx: integer): TObject;
  protected
    function CreateConnectionInfoObj: TObject; virtual;
    function BasicAddNode(NodeClass: TRakConnectableClass): TRakConnection;

    function CanConnect(Node: TRakConnectable): boolean; virtual;

    function DoCheckAutoDestroy: boolean; virtual;
    procedure DoDestroy; virtual;
    function DoConnect(Node: TRakConnectable): TRakConnection; overload; virtual;
    procedure DoDisconnect(Node: TRakConnectable); virtual;
    procedure DoDisconnectAll; virtual;
    function DoHasAnker: boolean; virtual;

    procedure ExcludeState(State: TRakConnectableStates); virtual;
    procedure IncludeState(State: TRakConnectableStates); virtual;

    function GetConnectionColClass: TRakConnectionColClass; virtual;

    function GetState: TRakConnectableState; virtual;

    function OnlyUnique: boolean; virtual;
  public
    constructor Create; virtual;
    destructor Destroy; override; //Statt Destroy sollte DoDestroy berschrieben werden

    function AcceptsNode(Node: TRakConnectable): boolean; virtual;
    procedure CheckAutoDestroy;

    function Connect(Node: TRakConnectable): TRakConnection; overload; virtual;
    procedure Disconnect(Node: TRakConnectable); overload;
    procedure DisconnectAll;

    function GetFilteredList(NotConnectable: TRakConnectable): TRakConnectableList; overload; virtual;
    function GetFilteredList(NotList: TRakConnectableList): TRakConnectableList; overload; virtual;
    function HasAnker: boolean; virtual;

    function IndexOfConnection(Connection: TRakConnection): integer;
    function IndexOfNode(Connection: TRakConnectable): integer;

    procedure PlugOff(Connection: TRakConnection; NodeNr: integer); virtual;
    procedure PlugOn(Connection: TRakConnection; NodeNr: integer); virtual;

    property AutoDestroy: boolean read FAutoDestroy write FAutoDestroy;
    property Count: integer read GetCount;
    property Connections[Idx: integer]: TRakConnection read GetConnections;
    property ConnectionInfoObjs[Idx: integer]: TObject read GetConnectionInfoObjs;
    property Nodes[Idx: integer]: TRakConnectable read GetNodes;
    property State: TRakConnectableState read GetState;
  end;

  TRakConnectableDestroy = class(TRakConnectable)
  private
    FOnDestroyed: TNotifyEvent;
  protected
    procedure DoDestroy; override;
  public
    property OnDestroyed: TNotifyEvent read FOnDestroyed write FOnDestroyed;
  end;

  TRakConnection = class(TRakConnectable)
  private
    FNode1: TRakConnectable;
    FNode2: TRakConnectable;
    FNode1InfoObj: TObject;
    FNode2InfoObj: TObject;

    function GetNode1InfoObj: TObject;
    function GetNode2InfoObj: TObject;
    procedure SetNode1InfoObj(const Value: TObject);
    procedure SetNode2InfoObj(const Value: TObject);
  protected
    procedure DoDestroy; override;

    procedure SetNode1(Value: TRakConnectable); virtual;
    procedure SetNode2(Value: TRakConnectable); virtual;
  public
    constructor Create; override;

    function ConnectedWith(KnownNode: TRakConnectable): TRakConnectable;
    function GetInfoObj(Node: TRakConnectable): TObject;

    property Node1: TRakConnectable read FNode1 write SetNode1;
    property Node2: TRakConnectable read FNode2 write SetNode2;
    property Node1InfoObj: TObject read GetNode1InfoObj write SetNode1InfoObj;
    property Node2InfoObj: TObject read GetNode2InfoObj write SetNode2InfoObj;
  end;

  TRakConnectionColItem = class(TRakCollectionItem)
  public
    Connection: TRakConnection;
  end;

  TRakConnectionCol = class(TRakCollection)
  private
    function GetItems(Idx: integer): TRakConnectionColItem;
  protected
    FParent: TRakConnectable;
  public
    function Add: TRakConnectionColItem;

    function ItemByConnection(Connection: TRakConnection): TRakConnectionColItem;
    function ItemByNode(Node: TRakConnectable): TRakConnectionColItem;

    property Items[Idx: integer]: TRakConnectionColItem read GetItems; default;
  end;

  TRakConnectableList = class(TList)
  private
    function GetItems(Idx: integer): TRakConnectable;
    procedure SetItems(Idx: integer; const Value: TRakConnectable);
  public
    property Items[Idx: integer]: TRakConnectable read GetItems write SetItems; default;
  end;

  TRakRollen = class;
  TRakRelationInfoObj = class;
  TBasisRolle = class end;
  TRolle = class of TBasisRolle;
//  TRolleBeispiel = class(TBasisRolle);

  TRakRelationClass = class of TRakRelation;

  TRakRelation =  class(TRakConnectable)
  private
    function GetConnectionInfoObjs(Idx: integer): TRakRelationInfoObj;
  protected
    FRollen: TRakRollen;

    function CreateConnectionInfoObj: TObject; override;
    procedure FillRollen; virtual; abstract;
    function RolleOK(Rolle: TRolle; EntityTyp: TClass): boolean;

    function DoConnect(Node: TRakConnectable; Rolle: TRolle): TRakConnection; overload; virtual;

    procedure DoDestroy; override;
  public
    constructor Create; override;

    function Connect(Node: TRakConnectable): TRakConnection; override;
    function Connect(Node: TRakConnectable; Rolle: TRolle): TRakConnection; overload; virtual;

    function GetConnection(Rolle: TRolle): TRakConnection; overload; virtual;
    function GetNode(Rolle: TRolle): TRakConnectable; overload; virtual;
    function GetRolle(Entity: TRakConnectable): TRolle; overload; virtual;

    property ConnectionInfoObjs[Idx: integer]: TRakRelationInfoObj read GetConnectionInfoObjs;
  end;

  TRakRelationInfoObj = class
  protected
    Rolle: TRolle;
  end;

  TRakRollenItem = class(TRakCollectionItem)
  public
    Rolle: TRolle;
    EntityTyp: TRakConnectableClass;
  end;

  TRakRollen = class(TRakCollection)
  private
    function GetItems(Idx: integer): TRakRollenItem;
  public
    function Add: TRakRollenItem;

    property Items[Idx: integer]: TRakRollenItem read GetItems; default;
  end;

  TRakZweierRelationClass = class of TRakZweierRelation;
  TRakZweierRelation = class(TRakRelation)
  private
    function GetEntity1: TRakConnectable;
    function GetEntity2: TRakConnectable;
  protected
    function CanConnect(Node: TRakConnectable): boolean; override;
  public
    function GetOtherEntity(Entity: TRakConnectable): TRakConnectable;

    property Entity1: TRakConnectable read GetEntity1;
    property Entity2: TRakConnectable read GetEntity2;
  end;


  TRakZweierRelationen = class(TRakConnectable)
  private
    function GetEntities(Idx: integer): TRakConnectable;
  protected
    FRelationClass: TRakZweierRelationClass;
    FRolle: TRolle;

    function ConnectEntity(Entity: TRakConnectable;
      RelationClass: TRakZweierRelationClass; EntityRolle: TRolle;
      SelfRolle: TRolle = nil): TRakZweierRelation; overload; virtual;

    procedure SetRelationClassAndRolle; virtual; abstract;
  public
    constructor Create; override;

    function ConnectEntity(Entity: TRakConnectable;
      Rolle: TRolle): TRakZweierRelation; overload; virtual;
    function GetRelation(Entity: TRakConnectable): TRakZweierRelation; virtual;

    property Entities[Idx: integer]: TRakConnectable read GetEntities;
  end;


implementation


{ TRakConnectable }


constructor TRakConnectable.Create;
begin
  inherited Create;

  FrcsWorking := 0;
  FAutoDestroy := false;
  FState := [];
  FConnections := GetConnectionColClass.Create;
  FConnections.FParent := self;
end;

destructor TRakConnectable.Destroy;
begin
  IncludeState(rcsDestroying);
  DoDestroy;
  FConnections.Free;

  inherited Destroy;
end;

procedure TRakConnectable.DoDestroy;
begin
  DisconnectAll;
end;

function TRakConnectable.BasicAddNode(NodeClass: TRakConnectableClass): TRakConnection;
var
  Node: TRakConnectable;
begin
  Node := NodeClass.Create;
  try
    Node.AutoDestroy := true;
    Result := Connect(Node);
    if Result = nil then
      Node.Free;
  except
    Node.Free;
    raise;
  end;
end;

procedure TRakConnectable.ExcludeState(State: TRakConnectableStates);
begin
  if State <> rcsWorking then
    FState := FState - [State];
  if FrcsWorking > 0 then
    dec(FrcsWorking);
end;

procedure TRakConnectable.IncludeState(State: TRakConnectableStates);
begin
  if State <> rcsWorking then
    FState := FState + [State];
  inc(FrcsWorking);
end;

function TRakConnectable.GetConnectionColClass: TRakConnectionColClass;
begin
  Result := TRakConnectionCol;
end;

function TRakConnectable.GetCount: integer;
begin
  Result := FConnections.Count;
end;

function TRakConnectable.GetConnections(Idx: integer): TRakConnection;
begin
  Result := FConnections[Idx].Connection;
end;

function TRakConnectable.GetNodes(Idx: integer): TRakConnectable;
var
  Con: TRakConnection;
begin
  Con := FConnections[Idx].Connection;
  Result := Con.ConnectedWith(self);
end;

function TRakConnectable.DoHasAnker: boolean;
var
  i: integer;
begin
  Result := false;
  i := 0;
  while not Result and (i < Count) do
  begin
    Result := Nodes[i].HasAnker;
    inc(i);
  end;
end;

function TRakConnectable.HasAnker: boolean;
begin
  Result := not AutoDestroy;
  if not Result then
  begin
    Result := not (rcsSearchingAnker in State);
    if Result then
    begin
      IncludeState(rcsSearchingAnker);
      try
        Result := DoHasAnker;
      finally
        ExcludeState(rcsSearchingAnker);
      end;
    end;
  end;
end;

function TRakConnectable.DoCheckAutoDestroy: boolean;
begin
  Result := not HasAnker;
end;

procedure TRakConnectable.CheckAutoDestroy;
var
  CanDestroy: boolean;
begin
  if AutoDestroy and (State = []) then
  begin
    IncludeState(rcsAutoDestroying);
    try
      CanDestroy := DoCheckAutoDestroy;
    finally
      ExcludeState(rcsAutoDestroying);
    end;

    if CanDestroy then
      Free;
  end;
end;

function TRakConnectable.Connect(Node: TRakConnectable): TRakConnection;
begin
  IncludeState(rcsConnecting);
  try
    Result := DoConnect(Node);
  finally
    ExcludeState(rcsConnecting);
  end;
end;

function TRakConnectable.GetState: TRakConnectableState;
begin
  if FrcsWorking > 0 then
    Result := FState + [rcsWorking]
  else
    Result := FState;
end;

function TRakConnectable.DoConnect(Node: TRakConnectable): TRakConnection;
var
  Item: TRakConnectionColItem;
  DoIt: boolean;
begin
  Result := nil;
  if Node <> nil then
  begin
    DoIt := true;
    if OnlyUnique then
      DoIt := Node.IndexOfNode(self) = -1;

    if DoIt and CanConnect(Node) then
    begin
      Item := FConnections.Add;
      Item.Connection := TRakConnection.Create;
      Result := Item.Connection;
      Result.Node1 := self;
      Result.Node2 := Node;
    end;
  end;
end;

procedure TRakConnectable.Disconnect(Node: TRakConnectable);
begin
  if Count > 0 then
  begin
    IncludeState(rcsDisconnecting);
    try
      DoDisconnect(Node);
    finally
      ExcludeState(rcsDisconnecting);
    end;

    CheckAutoDestroy;
  end;
end;

procedure TRakConnectable.DoDisconnect(Node: TRakConnectable);
var
  Con: TRakConnection;
  Item: TRakConnectionColItem;
begin
  repeat
    Item := FConnections.ItemByNode(Node);
    if Item <> nil then
    begin
      Con := Item.Connection;
      Item.Free;
      if (Con <> nil) and not (rcsDestroying in Con.State) then
        Con.Free;
    end;
  until Item = nil;
end;

procedure TRakConnectable.DisconnectAll;
begin
  if Count > 0 then
  begin
    IncludeState(rcsDisconnectingAll);
    try
      DoDisconnectAll;
    finally
      ExcludeState(rcsDisconnectingAll);
    end;

    CheckAutoDestroy;
  end;
end;

procedure TRakConnectable.DoDisconnectAll;
begin
  while Count > 0 do
    Disconnect(Nodes[0]);
end;

function TRakConnectable.IndexOfConnection(Connection: TRakConnection): integer;
var
  Item: TRakConnectionColItem;
begin
  Item := FConnections.ItemByConnection(Connection);
  if Item = nil then
    Result := -1
  else
    Result := Item.Index;
end;

function TRakConnectable.IndexOfNode(Connection: TRakConnectable): integer;
var
  Item: TRakConnectionColItem;
begin
  Item := FConnections.ItemByNode(Connection);
  if Item = nil then
    Result := -1
  else
    Result := Item.Index;
end;

procedure TRakConnectable.PlugOff(Connection: TRakConnection; NodeNr: integer);
var
  Idx: integer;
begin
  if Count > 0 then
  begin
    IncludeState(rcsDisconnecting);
    try
      Idx := IndexOfConnection(Connection);
      if Idx >= 0 then
      begin
        if  ((NodeNr = 1) and (Connection.Node1 = self) and (Connection.Node2 <> self))
            or
            ((NodeNr = 2) and (Connection.Node2 = self) and (Connection.Node1 <> self)) then
        begin
          FConnections[Idx].Free;
        end;
      end;
    finally
      ExcludeState(rcsDisconnecting);
    end;

    CheckAutoDestroy;
  end;
end;

procedure TRakConnectable.PlugOn(Connection: TRakConnection; NodeNr: integer);
var
  Item: TRakConnectionColItem;
begin
  if Connection <> nil then
  begin
    IncludeState(rcsConnecting);
    try
      if (NodeNr = 1) and (Connection.Node1 <> self) then
        Connection.Node1 := self
      else if (NodeNr = 2) and (Connection.Node2 <> self) then
        Connection.Node2 := self
      else if IndexOfConnection(Connection) < 0 then
      begin
        Item := FConnections.Add;
        Item.Connection := Connection;
      end;
    finally
      ExcludeState(rcsConnecting);
    end;
  end;
end;

function TRakConnectable.CreateConnectionInfoObj: TObject;
begin
  Result := nil;
end;

function TRakConnectable.GetConnectionInfoObjs(Idx: integer): TObject;
begin
  Result := Connections[Idx].GetInfoObj(self);
end;

function TRakConnectable.AcceptsNode(Node: TRakConnectable): boolean;
begin
  Result := Node is TRakConnectable
end;

function TRakConnectable.CanConnect(Node: TRakConnectable): boolean;
begin
  Result:= AcceptsNode(Node);
end;

function TRakConnectable.OnlyUnique: boolean;
begin
  Result := false;
end;

function TRakConnectable.GetFilteredList(
  NotList: TRakConnectableList): TRakConnectableList;
var
  i: integer;
  Node: TRakConnectable;
begin
  Result := TRakConnectableList.Create;
  try
    for i := 0 to Count - 1 do
    begin
      Node := Nodes[i];
      if NotList.IndexOf(Node) = -1 then
        Result.Add(Node);
    end;
  except
    Result.Free;
    raise;
  end;
end;

function TRakConnectable.GetFilteredList(
  NotConnectable: TRakConnectable): TRakConnectableList;
var
  i: integer;
  Node: TRakConnectable;
begin
  Result := TRakConnectableList.Create;
  try
    for i := 0 to Count - 1 do
    begin
      Node := Nodes[i];
      if Node <> NotConnectable then
        Result.Add(Node);
    end;
  except
    Result.Free;
    raise;
  end;
end;

{ TRakConnection }

constructor TRakConnection.Create;
begin
  inherited Create;

  FNode1 := nil;
  FNode2 := nil;
end;

procedure TRakConnection.DoDestroy;
begin
  inherited DoDestroy;

  Node1 := nil;
  Node2 := nil;
end;

procedure TRakConnection.SetNode1(Value: TRakConnectable);
begin
  if (Node1 <> Value) and not (rcsChangingNode1 in State) then
  begin
    IncludeState(rcsChangingNode1);
    try
      if Node1 <> nil then
      begin
        if not (rcsDestroying in Node1.State) then
        begin
          Node1.PlugOff(self, 1);
        end;
      end;

      FNode1InfoObj.Free;
      FNode1InfoObj := nil;

      FNode1 := Value;

      if Node1 <> nil then
      begin
        Node1.PlugOn(self, 1);
      end;
    finally
      ExcludeState(rcsChangingNode1);
    end;
  end;
end;

procedure TRakConnection.SetNode2(Value: TRakConnectable);
begin
  if (Node2 <> Value) and not (rcsChangingNode2 in State) then
  begin
    IncludeState(rcsChangingNode2);
    try
      if Node2 <> nil then
      begin
        if not (rcsDestroying in Node2.State) then
        begin
          Node2.PlugOff(self, 2);
        end;
      end;

      FNode2InfoObj.Free;
      FNode2InfoObj := nil;

      FNode2 := Value;

      if Node2 <> nil then
      begin
        Node2.PlugOn(self, 2);
      end;
    finally
      ExcludeState(rcsChangingNode2);
    end;
  end;
end;

function TRakConnection.ConnectedWith(KnownNode: TRakConnectable): TRakConnectable;
begin
  if Node1 = KnownNode then
    Result := Node2
  else if Node2 = KnownNode then
    Result := Node1
  else
    Result := nil;
end;




{procedure TRakConnection.SetNode1Rolle(const Value: TRolle);
begin
  FNode1Rolle := Value;
end;

procedure TRakConnection.SetNode2Rolle(const Value: TRolle);
begin
  FNode2Rolle := Value;
end;

function TRakConnection.NodeMitRolle(Rolle: TRolle): TRakConnectable;
begin
  if Node1Rolle.InheritsFrom(Rolle) and Node2Rolle.InheritsFrom(Rolle) then
    Result := nil
  else if Node1Rolle.InheritsFrom(Rolle) then
    Result := Node1
  else if Node2Rolle.InheritsFrom(Rolle) then
    Result := Node2
  else
    Result := nil
end;

procedure TRakConnection.SetRollen(Node: TRakConnectable; NodeRolle,
  AndereRolle: TRolle);
begin
  if Node1 = Node then
  begin
    Node1Rolle := NodeRolle;
    Node2Rolle := AndereRolle;
  end
  else if Node2 = Node then
  begin
    Node2Rolle := NodeRolle;
    Node1Rolle := AndereRolle;
  end;
end;}

function TRakConnection.GetNode1InfoObj: TObject;
begin
  if (FNode1InfoObj = nil) and (Node1 <> nil) then
    FNode1InfoObj := Node1.CreateConnectionInfoObj;

  Result := FNode1InfoObj;
end;

function TRakConnection.GetNode2InfoObj: TObject;
begin
  if (FNode2InfoObj = nil) and (Node2 <> nil) then
    FNode2InfoObj := Node1.CreateConnectionInfoObj;

  Result := FNode2InfoObj;
end;

function TRakConnection.GetInfoObj(Node: TRakConnectable): TObject;
begin
  if Node = Node1 then
    Result := Node1InfoObj
  else if Node = Node2 then
    Result := Node2InfoObj
  else
    Result := nil;
end;

procedure TRakConnection.SetNode1InfoObj(const Value: TObject);
begin
  if FNode1InfoObj = nil then
    FNode1InfoObj := Value
  else
    Value.Free;
end;

procedure TRakConnection.SetNode2InfoObj(const Value: TObject);
begin
  if FNode2InfoObj = nil then
    FNode2InfoObj := Value
  else
    Value.Free;
end;

{ TRakConnectionCol }

function TRakConnectionCol.GetItems(Idx: integer): TRakConnectionColItem;
begin
  Result := TRakConnectionColItem(inherited Items[Idx]);
end;

function TRakConnectionCol.Add: TRakConnectionColItem;
begin
  Result := TRakConnectionColItem(BasicAdd(TRakConnectionColItem));
end;

function TRakConnectionCol.ItemByConnection(Connection: TRakConnection): TRakConnectionColItem;
var
  i: integer;
begin
  Result := nil;
  i := 0;
  while (Result = nil) and (i < Count) do
  begin
    if Items[i].Connection = Connection then
      Result := Items[i];
    inc(i);
  end;
end;

function TRakConnectionCol.ItemByNode(Node: TRakConnectable): TRakConnectionColItem;
var
  i: integer;
begin
  Result := nil;
  i := 0;
  while (Result = nil) and (i < Count) do
  begin
    if Items[i].Connection.ConnectedWith(FParent) = Node then
      Result := Items[i];
    inc(i);
  end;
end;

{
function TRakConnectionCol.RollenCount(Rolle1, Rolle2: TRolle): integer;
var
  i: integer;
  Con: TRakConnection;
begin
  Result := 0;
  for i := 0 to Count -1 do
  begin
    Con := Items[i].Connection;
    if  (Con.Node1Rolle.InheritsFrom(Rolle1) and Con.Node2Rolle.InheritsFrom(Rolle2))
        or
        (Con.Node2Rolle.InheritsFrom(Rolle1) and Con.Node1Rolle.InheritsFrom(Rolle2)) then
    begin
      inc(Result);
    end;
  end;
end;

function TRakConnectionCol.RollenCount(Rolle: TRolle): integer;
var
  i: integer;
  Con: TRakConnection;
begin
  Result := 0;
  for i := 0 to Count -1 do
  begin
    Con := Items[i].Connection;
    if Con.Node1Rolle.InheritsFrom(Rolle) or Con.Node2Rolle.InheritsFrom(Rolle) then
      inc(Result);
  end;
end;

function TRakConnectionCol.RollenCount(Node: TRakConnectable;
  Rolle: TRolle): integer;
var
  i: integer;
  Con: TRakConnection;
begin
  Result := 0;
  for i := 0 to Count -1 do
  begin
    Con := Items[i].Connection;
    if  (Con.Node1Rolle.InheritsFrom(Rolle) and (Node = Con.Node2))
        or
        (Con.Node2Rolle.InheritsFrom(Rolle) and (Node = Con.Node1)) then
    begin
      inc(Result);
    end;
  end;
end;

function TRakConnectionCol.GetRollenItem(
  Rolle: TRolle; Idx: integer): TRakConnectionColItem;
var
  i, Counter: integer;
  Con: TRakConnection;
begin
  Result := nil;
  Counter := 0;
  for i := 0 to Count -1 do
  begin
    Con := Items[i].Connection;
    if Con.Node1Rolle.InheritsFrom(Rolle) or Con.Node2Rolle.InheritsFrom(Rolle) then
    begin
      if Counter = Idx then
      begin
        Result := Items[i];
        break;
      end;
      inc(Counter);
    end;
  end;
end;

function TRakConnectionCol.GetRollenItem(Node: TRakConnectable; Rolle: TRolle;
  Idx: integer): TRakConnectionColItem;
var
  Counter, i: integer;
  Con: TRakConnection;
begin
  Result := nil;

  Counter := 0;
  for i := 0 to Count -1 do
  begin
    Con := Items[i].Connection;
    if  (Con.Node1Rolle.InheritsFrom(Rolle) and (Node = Con.Node2))
        or
        (Con.Node2Rolle.InheritsFrom(Rolle) and (Node = Con.Node1)) then
    begin
      if Counter = Idx then
      begin
        Result := Items[i];
        break;
      end;
      inc(Counter);
    end;
  end;
end;}

{ TRakRelation }

constructor TRakRelation.Create;
begin
  inherited;

  FRollen := TRakRollen.Create;
  FillRollen;
  FAutoDestroy := true;
end;

function TRakRelation.DoConnect(Node: TRakConnectable;
  Rolle: TRolle): TRakConnection;
var
  Con: TRakConnection;
begin
  Result := nil;
  if Node <> nil then
  begin
    Con := GetConnection(Rolle);
    if Con = nil then
    begin
      if RolleOK(Rolle, Node.ClassType) then
      begin
        Result := inherited DoConnect(Node);
        if Result <> nil then
          TRakRelationInfoObj(Result.GetInfoObj(self)).Rolle := Rolle;
      end;
    end;
  end;
end;

procedure TRakRelation.DoDestroy;
begin
   FRollen.Free;
  FRollen := nil;

  inherited;
end;

function TRakRelation.RolleOK(Rolle: TRolle; EntityTyp: TClass): boolean;
var
  i: integer;
begin
  Result := false;
  for i := 0 to FRollen.Count - 1 do
  begin
    if (FRollen[i].Rolle = Rolle) and EntityTyp.InheritsFrom(FRollen[i].EntityTyp) then
    begin
      Result := true;
      break;
    end;
  end;
end;

function TRakRelation.GetConnection(Rolle: TRolle): TRakConnection;
var
  i: integer;
begin
  Result := nil;
  for i := 0 to Count - 1 do
  begin
    if ConnectionInfoObjs[i].Rolle.InheritsFrom(Rolle) then
    begin
      Result := Connections[i];
      break;
    end;
  end;
end;

function TRakRelation.Connect(Node: TRakConnectable): TRakConnection;
begin
  raise Exception.Create('Connect(TRakConnectable) darf nicht aufgerufen werden');
end;

function TRakRelation.Connect(Node: TRakConnectable;
  Rolle: TRolle): TRakConnection;
begin
  IncludeState(rcsConnecting);
  try
    Result := DoConnect(Node, Rolle);
  finally
    ExcludeState(rcsConnecting);
  end;
end;

function TRakRelation.GetNode(Rolle: TRolle): TRakConnectable;
var
  Con: TRakConnection;
begin
  Result := nil;
  Con := GetConnection(Rolle);
  if Con <> nil then
    Result := Con.ConnectedWith(self);
end;

function TRakRelation.CreateConnectionInfoObj: TObject;
begin
  Result := TRakRelationInfoObj.Create;
end;

function TRakRelation.GetConnectionInfoObjs(
  Idx: integer): TRakRelationInfoObj;
begin
  Result := TRakRelationInfoObj(inherited ConnectionInfoObjs[Idx]);
end;

function TRakRelation.GetRolle(Entity: TRakConnectable): TRolle;
var
  i: integer;
begin
  Result := nil;
  i := IndexOfNode(Entity);
  if i >= 0 then
    Result := ConnectionInfoObjs[i].Rolle;
end;

{ TRakRollen }

function TRakRollen.Add: TRakRollenItem;
begin
  Result := TRakRollenItem(BasicAdd(TRakRollenItem));
end;

function TRakRollen.GetItems(Idx: integer): TRakRollenItem;
begin
  Result := TRakRollenItem(inherited Items[Idx]);
end;


{ TRakZweierRelationen }

function TRakZweierRelationen.ConnectEntity(
  Entity: TRakConnectable; Rolle: TRolle): TRakZweierRelation;
begin
  Result := FRelationClass.Create;
  try
    if Result.Connect(self, FRolle) = nil then
    begin
      Result.Free;
      Result := nil;
    end
    else if Result.Connect(Entity, Rolle) = nil then
    begin
      Result.Free;
      Result := nil;
    end;
  except
    Result.Free;
    raise;
  end;
end;

function TRakZweierRelationen.ConnectEntity(Entity: TRakConnectable;
  RelationClass: TRakZweierRelationClass;
  EntityRolle, SelfRolle: TRolle): TRakZweierRelation;
begin
  Result := nil;
  if RelationClass.InheritsFrom(FRelationClass) then
  begin
    Result := RelationClass.Create;
    try
      if Result.Connect(self, SelfRolle) = nil then
      begin
        Result.Free;
        Result := nil;
      end
      else if Result.Connect(Entity, EntityRolle) = nil then
      begin
        Result.Free;
        Result := nil;
      end;
    except
      Result.Free;
      raise;
    end;
  end;
end;

constructor TRakZweierRelationen.Create;
begin
  inherited;

  SetRelationClassAndRolle;
end;

function TRakZweierRelationen.GetEntities(Idx: integer): TRakConnectable;
var
  Relation: TRakZweierRelation;
  Node: TRakConnectable;
begin
  Result := nil;
  Node := Nodes[Idx];
  if Node is TRakZweierRelation then
  begin
    Relation := TRakZweierRelation(Node);
    Result := Relation.GetOtherEntity(self);
  end;
end;

function TRakZweierRelationen.GetRelation(
  Entity: TRakConnectable): TRakZweierRelation;
var
  Relation: TRakZweierRelation;
  Node: TRakConnectable;
  Idx: integer;
begin
  Result := nil;
  for Idx := 0 to Count - 1 do
  begin
    Node := Nodes[Idx];
    if Node is TRakZweierRelation then
    begin
      Relation := TRakZweierRelation(Node);
      if Relation.GetOtherEntity(self) = Entity then
      begin
        Result := Relation;
        break;
      end;
    end;
  end;
end;

{ TRakZweierRelation }

function TRakZweierRelation.CanConnect(Node: TRakConnectable): boolean;
begin
  Result := (Count < 2) and inherited CanConnect(Node);
end;

function TRakZweierRelation.GetEntity1: TRakConnectable;
begin
  Result := nil;
  if Count > 0 then
    Result := Nodes[0];
end;

function TRakZweierRelation.GetEntity2: TRakConnectable;
begin
  Result := nil;
  if Count = 2 then
    Result := Nodes[1];
end;

function TRakZweierRelation.GetOtherEntity(
  Entity: TRakConnectable): TRakConnectable;
begin
  if Entity = Entity1 then
    Result := Entity2
  else if Entity = Entity2 then
    Result := Entity1
  else
    Result := nil;
end;

{ TRakConnectableList }

function TRakConnectableList.GetItems(Idx: integer): TRakConnectable;
begin
  Result := TRakConnectable(inherited Items[Idx]);
end;

procedure TRakConnectableList.SetItems(Idx: integer;
  const Value: TRakConnectable);
begin
  inherited;
end;

{ TRakConnectableDestroy }

procedure TRakConnectableDestroy.DoDestroy;
begin
  inherited;

  if Assigned(FOnDestroyed) then
    FOnDestroyed(self);
end;

initialization

finalization

end.
