unit RakPolyLine;
{///////////////////////////////////////////////////////////////////////////////
// Beschreibung : Die Unit RakPolyLine stellt die Daten-Klassen
                  TRakHorzVertLines zur Verfgung.
                  TRakHorzVertLines speichert die Positionen der Eckpunkte einer
                  Polyline, die nur aus waagerechten und senkrechten Strichen
                  besteht.
// erstellt am  : 4.1.2001
////////////////////////////////////////////////////////////////////////////////
// nderungen:
///////////////////////////////////////////////////////////////////////////////}
// Delphi 5 Implementation
//
// Copyright (c) 2000 by Richard Kaspar
// ("http://www.kasparsoft.de")
//
// Definitions:
// - "Package" refers to the collection of files distributed by
//   the Copyright Holder, and derivatives of that collection of
//   files created through textual modification.
// - "Standard Version" refers to such a Package if it has not
//   been modified, or has been modified in accordance with the
//   wishes of the Copyright Holder.
// - "Copyright Holder" is whoever is name in the copyright or
//   copyrights for the package.
// - "You" is you, if you're thinking about copying or distributing
//   this Package.
//
// Permission is hereby granted, free of charge, to any person
// obtaining a copy of this software and associated documentation
// files (the "Package"), to deal in the Package without restriction,
// including without limitation the rights to use, copy, modify,
// merge, publish, distribute, sublicense, and/or sell copies of the
// Package, and to permit persons to whom the Package is furnished
// to do so, subject to the following conditions:
//
// The above copyright notice and this permission notice shall be
// included in all copies or substantial portions of the Package.
//
// You may modify your copy of this Package in any way, provided
// that you insert a prominent notice in each changed file stating
// how and when you changed a file, and provided that you do at
// least one of the following:
//
// a) allow the Copyright Holder to include your modifications in
// the Standard Version of the Package.
//
// b) use the modified Package only within your corporation or
// organization.
//
// c) rename any non standard executables, units, and classes so
// the names do not conflict with standard executables, units, and
// classes, and provide a separate manual page that clearly documents
// how it differs from the standard version.
//
// d) make other distribution arrangements with the Copyright Holder.
//
// The name of the Copyright Holder may not be used to endorse or
// promote products derived from this Package without specific prior
// written permission.
//
// THE PACKAGE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
// EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
// MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.
// IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY
// CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT,
// TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE
// PACKAGE OR THE USE OR OTHER DEALINGS IN THE PACKAGE.
{$I RAK.INC}

interface

uses
  RakGlobalObjectID, RakUndoRedo,
  misctype, miscfunc, RakBinaryStreamData, RakDataObject;

type
{////////////////////////////////////////////////////////////////////////////////
// Klasse       : TRakLines
// Beschreibung : Mit TRakLines knnen die Eckpunkte einer Polyline
                  verwaltet werden.
// erstellt am  : 2.2.2003
////////////////////////////////////////////////////////////////////////////////
// nderungen:
////////////////////////////////////////////////////////////////////////////////}
  TRakLines = class(TRakDataObject)
  private
    FlclCount: integer;
    FlclPoints: TExtPointArr;

    procedure SetCount(const Value: integer);
    procedure ChangeFPoint(Idx: integer; APoint: TExtPoint);
    procedure CheckLength(Extra: integer = 0);

    property Count: integer read FlclCount write SetCount;

    function GetPoints(Idx: integer): TExtPoint;
    function GetPointsCount: integer;

    procedure DeletePoints(Idx, Anz: integer);
    procedure InsertPoints(Idx, Anz: integer);
  protected
    procedure CreateNow; override;
  public
    // Mit Assign knnen die Daten einer anderen Polyline bernommen werden
    procedure Assign(Source: TRakDataObject); override;

    // SaveToBinaryTag speichert die persistenten Daten in einem TRakBinaryTag
    procedure SaveToBinaryTag(Tag: TRakBinaryTag); override;
    procedure LoadFromBinaryTag(Tag: TRakBinaryTag); override;

    // Mit AddPoint kann ein neuer Eckpunkt hinzugefgt werden.
    procedure AddPoint(const NewPoint: TExtPoint);
    // Mit InsertPoint kann ein neuer Eckpunkt eingefgt werden.
    procedure InsertPoint(const Idx: integer; const NewPoint: TExtPoint);
    // Mit DeletePoint kann gezielt ein bestimmter Eckpunkt gelscht werden.
    procedure DeletePoint(const Idx: integer);
    // Lscht alle Linien
    procedure Clear;
    // Pack sorgt dafr, dass Linien der Lnge 0 aus der Polyline entfernt werden
    procedure Pack(von, bis: integer);
    // Mit Mirror kann man die Polyline an einer X- und Y-Achse spiegeln
    procedure Mirror(Horizontal, Vertical: boolean; XCenter, YCenter: extended);

    // Mit GetBounds kann man das Rechteck der maximalen Ausdehnung ermitteln
    procedure GetBounds(out Links, Oben, Breite, Hoehe: extended);

    // Mit Cut kann eine Linie in zwei gleich lange Linien aufgetrennt werden
    procedure Cut(Idx: integer);
    // GetPointIdxAt ermittelt welcher Eckpunkt sich an einer bestimmten Position befindet
    function GetPointIdxAt(X, Y: extended): integer;
    // Move verschiebt die ganze Polyline
    procedure Move(OffsetX, OffsetY: extended);
    // MoveToRect verschiebt und skalliert die Polyline so, dass sie ein bestimmtes Rechteck optimal ausfllt
    procedure MoveToRect(DestRect: TExtRect); virtual;

    // PointsCount gibt an, wieviele Eckpunkte die Polyline hat (PointsCount = LaengenCount + 1)
    property PointsCount: integer read GetPointsCount;
    // Points sind die einzelnen Eckpunkte der Polyline
    property Points[Idx: integer]: TExtPoint read GetPoints write ChangeFPoint; default;
  end;

{////////////////////////////////////////////////////////////////////////////////
// Klasse       : TRakHorzVertLines
// Beschreibung : Mit TRakHorzVertLines knnen die Eckpunkte einer Polyline
                  verwaltet werden, die nur aus senkrechten und waagerechten
                  Strichen besteht.
// erstellt am  : 4.1.2002
////////////////////////////////////////////////////////////////////////////////
// nderungen:
////////////////////////////////////////////////////////////////////////////////}
  TRakHorzVertLines = class(TRakDataObject)
  private
    FlclCount: integer;
    FlclPunkte: TExtPointArr;
    FlclStartHorz: boolean;

    function GetPunkte(Idx: integer): TExtPoint;
    procedure SetCount(const Value: integer);
    procedure SetStartHorz(const Value: boolean);
  private
    procedure ChangeFPunktX(Idx: integer; X: extended);
    procedure ChangeFPunktY(Idx: integer; Y: extended);
    procedure ChangeFPunkt(Idx: integer; Punkt: TExtPoint);
    procedure ChangeFStartHorz;
    procedure CheckLength(Extra: integer = 0);

    property Count: integer read FlclCount write SetCount;
    property Punkte[Idx: integer]: TExtPoint read GetPunkte write ChangeFPunkt;
    property StartHorz: boolean read FlclStartHorz write SetStartHorz;

    function GetLaengen(Idx: integer): extended;
    procedure SetLaengen(Idx: integer; const Value: extended);
    procedure SetStartpunkt(const Value: TExtPoint);
    function GetLaengenCount: integer;
    function GetPoints(Idx: integer): TExtPoint;
    function GetPointsCount: integer;
    function GetStartpunkt: TExtPoint;

    procedure DeletePoints(Idx, Anz: integer);
    procedure InsertPoints(Idx, Anz: integer);
    function GetStartPunktIstHorizontal: boolean;
  protected
    procedure CreateNow; override;
  public
    // Mit Assign knnen die Daten einer anderen Polyline bernommen werden
    procedure Assign(Source: TRakDataObject); override;

    // SaveToBinaryTag speichert die persistenten Daten in einem TRakBinaryTag
    procedure SaveToBinaryTag(Tag: TRakBinaryTag); override;
    procedure LoadFromBinaryTag(Tag: TRakBinaryTag); override;

    // Mit Add kann eine neue Linie hinzugefgt werden
    procedure Add(Laenge: extended);

    // Mit InsertPunkt kann ein neuer Eckpunkt hinzugefgt werden. Dies hat in
    // der Regel zur Folge, dass die Polyline danach zwei Linien mehr hat
    procedure InsertPunkt(const Idx: integer; const NewPoint: TExtPoint);
    // Mit DeletePunkt kann gezielt ein bestimmter Eckpunkt gelscht werden.
    // Dies hat in der Regel zur Folge, dass andere Eckpunkte mit gelscht oder
    // verschoben werden, damit die Linien weiterhin abwechselnd senkrecht waagerecht sind
    procedure DeletePunkt(const Idx: integer);
    // Lscht alle Linien
    procedure Clear;
    // Mit MovePoint kann ein Punkt auf eine bestimmte Position gendert werden,
    // Die beiden benachbarten Punkte werden evtl. auch mit verschoben, damit
    // der Polylinekarakter erhalten bleibt
    procedure MovePunkt(Idx: integer; NewPosition: TExtPoint);
    // Mit MoveLine kann man eine Linie verschieben
    procedure MoveLinie(Idx: integer; NewPosition: extended);
    // Pack sorgt dafr, dass Linien der Lnge 0 aus der Polyline entfernt werden
    procedure Pack(von, bis: integer);
    // Mit Mirror kann man die Polyline an einer X- und Y-Achse spiegeln
    procedure Mirror(Horizontal, Vertical: boolean; XCenter, YCenter: extended);

    // Mit GetBounds kann man das Rechteck der maximalen Ausdehnung ermitteln
    procedure GetBounds(out Links, Oben, Breite, Hoehe: extended);

    // Mit Trenne kann eine Linie in zwei Hlften aufgetrennt werden, die
    // eine neue Linie der Lnge 0 verbunden sind
    procedure Trenne(Idx: integer);
    procedure Verbinde(Idx: integer);
    // GetNodeIdxAt ermittelt welcher Eckpunkt sich an einer bestimmten Position befindet
    function GetNodeIdxAt(X, Y: extended): integer;
    // IsHorz stellt fest, ob eine bestimmte Linie horizontal oder vertikal ist
    function IsHorz(LaengeIdx: integer): boolean;
    // Move verschiebt die ganze Polyline
    procedure Move(OffsetX, OffsetY: extended);
    // MoveToRect verschiebt und skalliert die Polyline so, dass sie ein bestimmtes Rechteck optimal ausfllt
    procedure MoveToRect(DestRect: TExtRect); virtual;
    // Reset verndert die Polyline so, dass sie aus maximal drei Linien besteht, welche die beiden Punkte verbinden
    procedure Reset(Point1, Point2: TExtPoint); virtual;

    // An Startpunkt beginnt die Polyline
    property Startpunkt: TExtPoint read GetStartpunkt write SetStartpunkt;
    // StartPunktIstHorizontal gibt zurck, ob die erste Linie horizontal odeer vertikal ist
    property StartPunktIstHorizontal: boolean read GetStartPunktIstHorizontal;

    // LaengenCount gibt an, aus wievielen Linien die Polyline besteht
    property LaengenCount: integer read GetLaengenCount;
    // Laengen sind die die Lngen der einzelnen Linien
    property Laengen[Idx: integer]: extended read GetLaengen write SetLaengen; default;

    // PointsCoutn gibt an, wieviele Eckpunkte die Polyline hat (PointsCount = LaengenCount + 1)
    property PointsCount: integer read GetPointsCount;
    // Points sind die einzelnen Eckpunkte der Polyline
    property Points[Idx: integer]: TExtPoint read GetPoints;
  end;

////////////////////////////////////////////////////////////////////////////////
// Ab hier kommen die Klassen, die fr die Untersttzung des
// Undo/Redo-Mechanismuses bentigt werden

  TRakHvlHcFCount = class(TRakHcInteger)
  protected
    procedure SetValue(Obj: TObject; Value: integer; var Done: boolean); override;
  end;

  TRakHvlHcClear = class(TRakHcBinaryTag)
  protected
    function UndoNow: boolean; override;
    function RedoNow: boolean; override;
  end;

  TRakHvlHcFPunkt = class(TRakHcIdxTExtPoint)
  protected
    procedure SetValue(Obj: TObject; Value: TExtPoint; var Done: boolean); override;
  end;

  TRakHvlHcFPunktX = class(TRakHcIdxExtended)
  protected
    procedure SetValue(Obj: TObject; Value: extended; var Done: boolean); override;
  end;

  TRakHvlHcFPunktY = class(TRakHcIdxExtended)
  protected
    procedure SetValue(Obj: TObject; Value: extended; var Done: boolean); override;
  end;

  TRakHvlHcFStartHorz = class(TRakHcBoolean)
  protected
    procedure SetValue(Obj: TObject; Value: boolean; var Done: boolean); override;
  end;

  TRakHvlHcCheckLength = class(TRakHcInteger)
  protected
    procedure SetValue(Obj: TObject; Value: integer; var Done: boolean); override;
  end;

implementation

{ TRakHorzVertLines }

procedure TRakHorzVertLines.Add(Laenge: extended);
begin
  NewGroupStart(TRakCommandGroup.Create);
  try
    CheckLength(1);

    if IsHorz(Count-2) then
      Punkte[Count] := ExtPoint(Punkte[Count-1].X, Punkte[Count-1].Y + Laenge)
    else
      Punkte[Count] := ExtPoint(Punkte[Count-1].X + Laenge, Punkte[Count-1].Y);

    Count := Count + 1;
  finally
    NewGroupStop;
  end;
end;

procedure TRakHorzVertLines.Assign(Source: TRakDataObject);
var
  Src: TRakHorzVertLines;
  i: integer;
begin
  if Source is TRakHorzVertLines then
  begin
    Src := TRakHorzVertLines(Source);
    Count := Src.Count;
    CheckLength(0);
    for i := 0 to Count - 1 do
      Punkte[i] := Src.Punkte[i];
    StartHorz := Src.StartHorz;
  end;
end;

procedure TRakHorzVertLines.ChangeFPunkt(Idx: integer; Punkt: TExtPoint);
var
  Alt: TExtPoint;
begin
  if not ExtPointsEqual(FlclPunkte[Idx], Punkt) then
  begin
    Alt := FlclPunkte[Idx];
    FlclPunkte[Idx] := Punkt;
    AddCommand(TRakHvlHcFPunkt.Init(Idx, Alt, Punkt));
  end;
end;

procedure TRakHorzVertLines.ChangeFPunktX(Idx: integer; X: extended);
var
  Alt: extended;
begin
  if FlclPunkte[Idx].X <> X then
  begin
    Alt := FlclPunkte[Idx].X;
    FlclPunkte[Idx].X := X;
    AddCommand(TRakHvlHcFPunktX.Init(Idx, Alt, X));
  end;
end;

procedure TRakHorzVertLines.ChangeFPunktY(Idx: integer; Y: extended);
var
  Alt: extended;
begin
  if FlclPunkte[Idx].Y <> Y then
  begin
    Alt := FlclPunkte[Idx].Y;
    FlclPunkte[Idx].Y := Y;
    AddCommand(TRakHvlHcFPunktY.Init(Idx, Alt, Y));
  end;
end;

procedure TRakHorzVertLines.ChangeFStartHorz;
begin
  FlclStartHorz := not FlclStartHorz;
  AddCommand(TRakHvlHcFStartHorz.Init(FlclStartHorz));
end;

procedure TRakHorzVertLines.CheckLength(Extra: integer);
var
  Alt: integer;
begin
  if Count + Extra > Length(FlclPunkte) then
  begin
    Alt := length(FlclPunkte);
    SetLength(FlclPunkte, Count + Extra + 10);
    AddCommand(TRakHvlHcCheckLength.Init(Alt, length(FlclPunkte)));
  end;
end;

procedure TRakHorzVertLines.Clear;
var
  Tag: TRakBinaryTag;
begin
  Tag := GetAsBinaryTag;
  FlclCount := 0;
  AddCommand(TRakHvlHcClear.Init(Tag));
end;

procedure TRakHorzVertLines.CreateNow;
begin
  inherited;

  FlclCount := 2;
  CheckLength;
  FlclPunkte[0].X := 0;
  FlclPunkte[0].Y := 0;
  FlclPunkte[1].X := 0;
  FlclPunkte[1].Y := 10;
end;

procedure TRakHorzVertLines.DeletePoints(Idx, Anz: integer);
var
  i: integer;
begin
  if Idx < Count then
  begin
    if Idx + Anz > Count then
      Anz := Count - Idx;
    for i := Idx to Count - Anz - 1 do
      Punkte[i] := Punkte[i+Anz];

    Count := Count - Anz;
  end;
end;

procedure TRakHorzVertLines.DeletePunkt(const Idx: integer);
var
  Old: TExtPoint;
begin
  NewGroupStart(TRakCommandGroup.Create);
  try
    if IsBetween(Idx, 0, Count-1) then
    begin
      if Count < 3 then
      begin // Wenn wir hchstens eine Linie haben
        // Dann wird alles gelscht
        Clear;
      end
      else
      begin
        Old := Punkte[Idx];

        if (Idx = 0) or (Idx = Count-1) then
        begin // Wenn der erste oder der letzte Punkt gelscht werden soll
          DeletePoints(Idx, 1);
          if Idx = 0 then
            ChangeFStartHorz;
        end
        else if Idx = 1 then
        begin // Wenn der zweite Punkt gelscht werden soll
          DeletePoints(Idx, 1);
          if StartPunktIstHorizontal then
            ChangeFPunktX(Idx, Punkte[Idx-1].X)
          else
            ChangeFPunktY(Idx, Punkte[Idx-1].Y);
        end
        else if Idx = Count-2 then
        begin // Wenn der vorletzte Punkt gelscht werden soll
          DeletePoints(Idx, 1);
          if IsHorz(Idx) then
            ChangeFPunktY(Idx, Punkte[Idx+1].Y)
          else
            ChangeFPunktX(Idx, Punkte[Idx+1].X);
        end
        else
        begin
          DeletePoints(Idx, 2);
          if IsHorz(Idx) then
            ChangeFPunktX(Idx-1, Punkte[Idx].X)
          else
            ChangeFPunktY(Idx-1, Punkte[Idx].Y);
        end;
      end;
    end;
  finally
    NewGroupStop;
  end;
end;


procedure TRakHorzVertLines.GetBounds(out Links, Oben, Breite, Hoehe: extended);
var
  X, Y, Rechts, Unten: extended;
  i: integer;
begin
  X := Punkte[0].X;
  Y := Punkte[0].Y;
  Links := X;
  Oben := Y;
  Rechts := X;
  Unten := Y;
  for i := 1 to PointsCount - 1 do
  begin
    X := Punkte[i].X;
    Y := Punkte[i].Y;

    if X < Links then
      Links := X
    else if X > Rechts then
      Rechts := X;

    if Y < Oben then
      Oben := Y
    else if Y > Unten then
      Unten := Y;
  end;

  Breite := Rechts - Links;
  Hoehe := Unten - Oben;
end;

function TRakHorzVertLines.GetLaengen(Idx: integer): extended;
var
  X, Y: extended;
begin
  X := Punkte[Idx+1].X - Punkte[Idx].X;
  Y := Punkte[Idx+1].Y - Punkte[Idx].Y;
  if abs(X) < abs(Y) then
    Result := Y
  else
    Result := X;
end;

function TRakHorzVertLines.GetLaengenCount: integer;
begin
  Result := Count -1;
end;

function TRakHorzVertLines.GetNodeIdxAt(X, Y: extended): integer;
var
  i: integer;
begin
  Result := -1;
  for i := 0 to PointsCount - 1 do
  begin
    if (Punkte[i].X = X) and (Punkte[i].Y = Y) then
    begin
      Result := i;
      break;
    end;
  end;
end;

function TRakHorzVertLines.GetPoints(Idx: integer): TExtPoint;
begin
  Result := Punkte[Idx];
end;

function TRakHorzVertLines.GetPointsCount: integer;
begin
  Result := Count;
end;

function TRakHorzVertLines.GetPunkte(Idx: integer): TExtPoint;
begin
  Result := FlclPunkte[Idx];
end;

function TRakHorzVertLines.GetStartpunkt: TExtPoint;
begin
  Result := Punkte[0];
end;

function TRakHorzVertLines.GetStartPunktIstHorizontal: boolean;
begin
  Result := StartHorz;
end;

procedure TRakHorzVertLines.InsertPoints(Idx, Anz: integer);
var
  i: integer;
begin
  if Idx < Count then
  begin
    CheckLength(Anz);
    for i := Count + Anz - 1 downto Idx + 1 do
      Punkte[i] := Punkte[i-Anz];

    Count := Count + Anz;
  end;
end;

procedure TRakHorzVertLines.InsertPunkt(const Idx: integer; const NewPoint: TExtPoint);
begin
  NewGroupStart(TRakCommandGroup.Create);
  try
    if (Idx = 0) or (Idx = Count-1) then
    begin
      if ((Idx = 0) and IsHorz(Idx)) or ((Idx = Count-1) and IsHorz(Idx-1)) then
        ChangeFPunktX(Idx, NewPoint.X)
      else
        ChangeFPunktY(Idx, NewPoint.Y);
      InsertPoints(Idx, 1);
      if Idx = 0 then
      begin
        Punkte[Idx] := NewPoint;
        ChangeFStartHorz;
      end
      else
        Punkte[Idx+1] := NewPoint;
    end
    else
    begin
      InsertPoints(Idx, 2);
      Punkte[Idx+1] := NewPoint;
      if IsHorz(Idx) then
      begin
        ChangeFPunktY(Idx, NewPoint.Y);
        ChangeFPunktX(Idx+2, NewPoint.X);
      end
      else
      begin
        ChangeFPunktX(Idx, NewPoint.X);
        ChangeFPunktY(Idx+2, NewPoint.Y);
      end;
    end;
  finally
    NewGroupStop;
  end;
end;

function TRakHorzVertLines.IsHorz(LaengeIdx: integer): boolean;
begin
  Result := ((LaengeIdx mod 2) = 1) xor StartHorz;
end;

procedure TRakHorzVertLines.LoadFromBinaryTag(Tag: TRakBinaryTag);
var
  i: integer;
  List: IRakBinaryTagList;
  SubTag: TRakBinaryTag;
begin
  if Tag <> nil then
  begin
    List := Tag.FindTagsByName('Points');
    Count := List.Count;
    CheckLength(10);
    for i := 0 to Count - 1 do
    begin
      SubTag := List.Items(i);
      Punkte[i] := ExtPoint(SubTag.AttributeAsFloat('X', 0),
        SubTag.AttributeAsFloat('Y', 0));
    end;
    StartHorz := Tag.AttributeAsBoolean('Horz', true);
  end;
end;

procedure TRakHorzVertLines.Mirror(Horizontal, Vertical: boolean; XCenter, YCenter: extended);
var
  i: integer;
  APoint: TExtPoint;
begin
  if Horizontal or Vertical then
  begin
    NewGroupStart(TRakCommandGroup.Create);
    try
      for i := 0 to PointsCount - 1 do
      begin
        APoint := Punkte[i];

        if Horizontal then
          APoint.X := 2*XCenter - APoint.X;

        if Vertical then
          APoint.Y := 2*YCenter - APoint.Y;

        Punkte[i] := APoint;
      end;
    finally
      NewGroupStop;
    end;
  end;
end;

procedure TRakHorzVertLines.Move(OffsetX, OffsetY: extended);
var
  i: integer;
begin
  NewGroupStart(TRakCommandGroup.Create);
  try
    for i := 0 to PointsCount - 1 do
      Punkte[i] := ExtPoint(Punkte[i].X + OffsetX, Punkte[i].Y + OffsetY);
  finally
    NewGroupStop;
  end;
end;

procedure TRakHorzVertLines.MoveLinie(Idx: integer; NewPosition: extended);
var
  Horz: boolean;
  PunktA, PunktEnd: TExtPoint;
begin
  NewGroupStart(TRakCommandGroup.Create);
  try
    Horz := IsHorz(Idx);
    if (Horz and (NewPosition <> Punkte[Idx].Y)) or (not Horz and (NewPosition <> Punkte[Idx].X)) then
    begin
      // Merken, wo er vorher stand
      PunktA := Punkte[Idx];
      PunktEnd := Punkte[Idx+1];
      // Auf die neue Position setzen
      if Horz then
      begin
        ChangeFPunktY(Idx, NewPosition);
        ChangeFPunktY(Idx+1, NewPosition);
      end
      else
      begin
        ChangeFPunktX(Idx, NewPosition);
        ChangeFPunktX(Idx+1, NewPosition);
      end;

      if Idx = Count-2 then
      begin
        InsertPoints(Idx+1, 1);
        Punkte[Idx+2] := PunktEnd;
      end;

      if Idx = 0 then
      begin
        InsertPoints(0, 1);
        Punkte[0] := PunktA;
        ChangeFStartHorz;
      end;
      Pack(0, Count-1);
    end;
  finally
    NewGroupStop;
  end;
end;

procedure TRakHorzVertLines.MovePunkt(Idx: integer; NewPosition: TExtPoint);
var
  Horz: boolean;
  Punkt: TExtPoint;
begin
  NewGroupStart(TRakCommandGroup.Create);
  try
    if not ExtPointsEqual(NewPosition, Punkte[Idx]) then
    begin // Falls sich der Punkt wirklich verschiebt
      // Merken, wo er vorher stand
      Punkt := Punkte[Idx];
      // Auf die neue Position setzen
      Punkte[Idx] := NewPosition;
      //Feststellen, ob die entsprechende Linie horizontal ist
      if Idx = Count -1 then
        Horz := not IsHorz(Idx-1)
      else
        Horz := IsHorz(Idx);

      // Den unmittelbaren Nachbarpunkten ihre neue Position verpassen
      if Horz then
      begin
        if Idx > 0 then
          ChangeFPunktX(Idx-1, NewPosition.X);
        if Idx < Count - 1 then
          ChangeFPunktY(Idx+1, NewPosition.Y);
      end
      else
      begin
        if Idx < Count - 1 then
          ChangeFPunktX(Idx+1, NewPosition.X);
        if Idx > 0 then
          ChangeFPunktY(Idx-1, NewPosition.Y);
      end;

      // Wenn wir den zweiten oder vorletzten Punkt verschoben haben, dann mssen wir neue Punkte erzeugen,
      // damit sichergestellt ist, dass die Linien noch von den gleichen Punkten aus starten.
      // Nur wenn ein Endpunkte selbst verschoben wird, starten danach die Lininen von einem anderen Punkt
      if Idx = Count-2 then
      begin // Wenn wir den vorletzten Punkt verschoben haben
        if Horz then
          InsertPunkt(Idx+1, ExtPoint(Punkte[Idx+1].X, Punkt.Y))
        else
          InsertPunkt(Idx+1, ExtPoint(Punkt.X, Punkte[Idx+1].Y));
      end;
      if Idx = 1 then
      begin // Wenn wir den zweiten Punkt verschoben haben
        if Horz then
          InsertPunkt(0, ExtPoint(Punkt.X, Punkte[0].Y))
        else
          InsertPunkt(0, ExtPoint(Punkte[0].X, Punkt.Y));

        // Und zum Schluss noch bereinanderliegende Punkte entsorgen
        Pack(0, GetKleiner(4, Count-1));
      end
      else
        Pack(GetGroesser(Idx-2, 0), GetKleiner(Idx+2, Count-1));
    end;
  finally
    NewGroupStop;
  end;
end;

procedure TRakHorzVertLines.MoveToRect(DestRect: TExtRect);
var
  i: integer;
  Links, Oben, Breite, Hoehe: extended;
  XScale, YScale: extended;
  APoint: TExtPoint;
begin
  NewGroupStart(TRakCommandGroup.Create);
  try
    GetBounds(Links, Oben, Breite, Hoehe);
    DestRect := NormalizeRect(DestRect);

    if Breite = 0 then
      XScale := 0
    else
      XScale := (DestRect.Right - DestRect.Left) / Breite;

    if Hoehe = 0 then
      YScale := 0
    else
      YScale := (DestRect.Bottom - DestRect.Top) / Hoehe;

    for i := 0 to PointsCount - 1 do
    begin
      APoint := Punkte[i];
      APoint.X := (APoint.X - Links) * XScale + DestRect.Left;
      APoint.Y := (APoint.Y - Oben) * YScale + DestRect.Top;
      Punkte[i] := APoint;
    end;
  finally
    NewGroupStop;
  end;
end;

procedure TRakHorzVertLines.Pack(von, bis: integer);
var
  i: integer;
begin
  NewGroupStart(TRakCommandGroup.Create);
  try
    // Jetzt entsorgen wir noch Punkte, die aufeinander liegen
    if Count > 2 then
    begin
      KleinerGroesser(von, bis);

      // Der erste Punkt bentigt eine Sonderbahndlung
      while (von = 0) and (Count > 2) do
      begin // Solange der erste Punkt mit berprft werden soll und wir mehr als eine Linie haben
        if ExtPointsEqual(Punkte[0], Punkte[1]) then
        begin // Wenn der erste und zweite Punkt aufeinander liegen
          DeletePoints(0, 1); // Dann schmeien wir den ersten raus
          ChangeFStartHorz;
          dec(bis);
        end
        else // Wenn die Punkte verschieden sind
          inc(von); //Dann knnen wir mit dem zweiten weitermachen
      end;

      // Der letzte Punkt bentigt auch eine Sonderbahndlung
      while (bis = Count-1) and (Count > 2) do
      begin // Solange der letzte Punkt mit berprft werden soll und wir mehr als eine Linie haben
        if ExtPointsEqual(Punkte[bis], Punkte[bis-1]) then
        begin // Wenn der letzte und vorletzte Punkt aufeinander liegen
          DeletePoints(bis, 1) // Dann schmeien wir den letzten raus
        end;

        dec(bis);
      end;

      i := bis;
      while (i > von) and (Count > 2) do
      begin
        if ExtPointsEqual(Punkte[i], Punkte[i-1]) then
        begin
          DeletePoints(i-1, 2);
          dec(i, 2);
        end
        else
        begin
          dec(i);
        end;
      end;
    end;
  finally
    NewGroupStop;
  end;
end;
{
function TRakHorzVertLines.Redo(ChangeData: TRakCommand): boolean;
begin
  Result := true;
  if ChangeData is TRakHvlHcClear then
    Clear
  else if ChangeData is TRakHvlHcFPunkt then
    Punkte[TRakHvlHcFPunkt(ChangeData).Idx] := TRakHvlHcFPunkt(ChangeData).Neu
  else if ChangeData is TRakHvlHcFPunktX then
    ChangeFPunktX(TRakHvlHcFPunktX(ChangeData).Idx, TRakHvlHcFPunktX(ChangeData).Neu)
  else if ChangeData is TRakHvlHcFPunktY then
    ChangeFPunktY(TRakHvlHcFPunktY(ChangeData).Idx, TRakHvlHcFPunktY(ChangeData).Neu)
  else if ChangeData is TRakHvlHcFStartHorz then
    StartHorz := TRakHvlHcFStartHorz(ChangeData).Neu
  else if ChangeData is TRakHvlHcCheckLength then
    SetLength(FlclPunkte, TRakHvlHcCheckLength(ChangeData).Neu)
  else if ChangeData is TRakHvlHcFCount then
    Count := TRakHvlHcFCount(ChangeData).Neu
  else
    Result := inherited Redo(ChangeData);
end;
}
procedure TRakHorzVertLines.Reset(Point1, Point2: TExtPoint);
begin
  NewGroupStart(TRakCommandGroup.Create);
  try
    if (Point1.X = Point2.X) or (Point1.Y = Point2.Y) then
    begin
      Count := 2;
      Punkte[0] := Point1;
      Punkte[1] := Point2;
      StartHorz := Point1.Y = Point2.Y;
    end
    else
    begin
      Count := 4;
      CheckLength;
      Punkte[0] := Point1;
      Punkte[1] := Point1;
      Punkte[2] := Point2;
      Punkte[3] := Point2;
      if abs(Point1.X - Point2.X) > abs(Point1.Y - Point2.Y) then
      begin
        ChangeFPunktX(1, (Point1.X + Point2.X) / 2);
        ChangeFPunktX(2, Punkte[1].X);
        StartHorz := true;
      end
      else
      begin
        ChangeFPunktY(1, (Point1.Y + Point2.Y) / 2);
        ChangeFPunktY(2, Punkte[1].Y);
        StartHorz := false;
      end;
    end;
  finally
    NewGroupStop;
  end;
end;

procedure TRakHorzVertLines.SaveToBinaryTag(Tag: TRakBinaryTag);
var
  i: integer;
  SubTag: TRakBinaryTag;
begin
  for i := 0 to Count - 1 do
  begin
    SubTag := Tag.AddTag('Points');
    TRakBinaryExtended(SubTag.AddAttribute(TRakBinaryExtended, 'X')).Value := Punkte[i].X;
    TRakBinaryExtended(SubTag.AddAttribute(TRakBinaryExtended, 'Y')).Value := Punkte[i].Y;
  end;
  TRakBinaryBoolean(Tag.AddAttribute(TRakBinaryBoolean, 'Horz')).Value := StartHorz;
end;

procedure TRakHorzVertLines.SetCount(const Value: integer);
var
  Alt: integer;
begin
  if FlclCount <> Value then
  begin
    Alt := FlclCount;
    FlclCount := Value;
    AddCommand(TRakHvlHcFCount.Init(Alt, Value));
  end;
end;

procedure TRakHorzVertLines.SetLaengen(Idx: integer; const Value: extended);
begin
  if (0 <= Idx) and (Idx < Count-1) then
  begin
    if Punkte[Idx].X = Punkte[Idx+1].X then
      ChangeFPunktY(Idx+1, Punkte[Idx].Y + Value)
    else
      ChangeFPunktX(Idx+1, Punkte[Idx].X + Value);
  end;
end;

procedure TRakHorzVertLines.SetStartHorz(const Value: boolean);
begin
  if Value <> FlclStartHorz then
    ChangeFStartHorz;
end;

procedure TRakHorzVertLines.SetStartpunkt(const Value: TExtPoint);
begin
  if IsHorz(0) then
    ChangeFPunktY(1, Value.Y)
  else
    ChangeFPunktX(1, Value.X);

  Punkte[0] := Value;
end;

procedure TRakHorzVertLines.Trenne(Idx: integer);
begin
  NewGroupStart(TRakCommandGroup.Create);
  try
    if (0 <= Idx) and (Idx < Count-1) then
    begin
      CheckLength(2);

      InsertPoints(Idx+1, 2);

      Laengen[Idx] := Laengen[Idx] / 2;
      Laengen[Idx+1] := 0;
      Laengen[Idx+2] := Laengen[Idx];
    end;
  finally
    NewGroupStop;
  end;
end;
{
function TRakHorzVertLines.Undo(ChangeData: TRakCommand): boolean;
begin
  Result := true;
  if ChangeData is TRakHvlHcClear then
    LoadFromBinaryTag(TRakHvlHcClear(ChangeData).Tag)
  else if ChangeData is TRakHvlHcFPunkt then
    Punkte[TRakHvlHcFPunkt(ChangeData).Idx] := TRakHvlHcFPunkt(ChangeData).Alt
  else if ChangeData is TRakHvlHcFPunktX then
    ChangeFPunktX(TRakHvlHcFPunktX(ChangeData).Idx, TRakHvlHcFPunktX(ChangeData).Alt)
  else if ChangeData is TRakHvlHcFPunktY then
    ChangeFPunktY(TRakHvlHcFPunktY(ChangeData).Idx, TRakHvlHcFPunktY(ChangeData).Alt)
  else if ChangeData is TRakHvlHcFStartHorz then
    StartHorz := not TRakHvlHcFStartHorz(ChangeData).Neu
  else if ChangeData is TRakHvlHcCheckLength then
    SetLength(FlclPunkte, TRakHvlHcCheckLength(ChangeData).Alt)
  else if ChangeData is TRakHvlHcFCount then
    Count := TRakHvlHcFCount(ChangeData).Alt
  else
    Result := inherited Redo(ChangeData);
end;
}
procedure TRakHorzVertLines.Verbinde(Idx: integer);
begin
  NewGroupStart(TRakCommandGroup.Create);
  try
    if (0 <= Idx) and (Idx < Count-1) then
    begin
      if Idx = 0 then
      begin // Wenn die erste Linie verschwinden soll
        DeletePoints(0,1);
      end
      else if Idx = Count-2 then
      begin // Wenn die letzte Linie verschwinden soll
        DeletePoints(Idx+1,1);
      end
      else
      begin // Wenn zwei Linien mittendrin verschwinden sollen
        Laengen[Idx-1] := Laengen[Idx-1] + Laengen[Idx+1];
        DeletePoints(Idx+1,2);
      end;
    end;
  finally
    NewGroupStop;
  end;
end;

{ TRakHvlHcFCount }

procedure TRakHvlHcFCount.SetValue(Obj: TObject; Value: integer;
  var Done: boolean);
begin
  (Obj as TRakHorzVertLines).Count := Value;
end;

{ TRakHvlHcClear }

function TRakHvlHcClear.RedoNow: boolean;
var
  Obj: TObject;
begin
  Obj := ChangedObject;
  if Obj is TRakHorzVertLines then
  begin
    Result := true;
    TRakHorzVertLines(Obj).Clear;
  end
  else
    Result := false;
end;

function TRakHvlHcClear.UndoNow: boolean;
var
  Obj: TObject;
begin
  Obj := ChangedObject;
  if Obj is TRakHorzVertLines then
  begin
    Result := true;
    TRakHorzVertLines(Obj).LoadFromBinaryTag(FValue);
  end
  else
    Result := false;
end;

{ TRakHvlHcFPunkt }

procedure TRakHvlHcFPunkt.SetValue(Obj: TObject; Value: TExtPoint;
  var Done: boolean);
begin
  (Obj as TRakHorzVertLines).Punkte[Idx] := Value;
end;

{ TRakHvlHcFPunktX }

procedure TRakHvlHcFPunktX.SetValue(Obj: TObject; Value: extended;
  var Done: boolean);
begin
  (Obj as TRakHorzVertLines).ChangeFPunktX(Idx, Value);
end;

{ TRakHvlHcFPunktY }

procedure TRakHvlHcFPunktY.SetValue(Obj: TObject; Value: extended;
  var Done: boolean);
begin
  (Obj as TRakHorzVertLines).ChangeFPunktY(Idx, Value);
end;

{ TRakHvlHcFStartHorz }

procedure TRakHvlHcFStartHorz.SetValue(Obj: TObject; Value: boolean; var Done: boolean);
begin
  (Obj as TRakHorzVertLines).StartHorz := Value;
end;

{ TRakHvlHcCheckLength }

procedure TRakHvlHcCheckLength.SetValue(Obj: TObject; Value: integer;
  var Done: boolean);
begin
  SetLength((Obj as TRakHorzVertLines).FlclPunkte, Value)
end;

{ TRakLines }

procedure TRakLines.AddPoint(const NewPoint: TExtPoint);
begin
	InsertPoint(PointsCount, NewPoint);
end;

procedure TRakLines.Assign(Source: TRakDataObject);
var
  Src: TRakLines;
  i: integer;
begin
  if Source is TRakLines then
  begin
    Src := TRakLines(Source);
    Count := Src.Count;
    CheckLength(0);
    for i := 0 to Count - 1 do
      Points[i] := Src.Points[i];
  end;
end;

procedure TRakLines.ChangeFPoint(Idx: integer; APoint: TExtPoint);
var
  Alt: TExtPoint;
begin
  if not ExtPointsEqual(FlclPoints[Idx], APoint) then
  begin
    Alt := FlclPoints[Idx];
    FlclPoints[Idx] := APoint;
    AddCommand(TRakHvlHcFPunkt.Init(Idx, Alt, APoint));
  end;
end;

procedure TRakLines.CheckLength(Extra: integer);
var
  Alt: integer;
begin
  if Count + Extra > Length(FlclPoints) then
  begin
    Alt := length(FlclPoints);
    SetLength(FlclPoints, Count + Extra + 10);
    AddCommand(TRakHvlHcCheckLength.Init(Alt, length(FlclPoints)));
  end;
end;

procedure TRakLines.Clear;
var
  Tag: TRakBinaryTag;
begin
  Tag := GetAsBinaryTag;
  FlclCount := 0;
  AddCommand(TRakHvlHcClear.Init(Tag));
end;

procedure TRakLines.CreateNow;
begin
  inherited;

  FlclCount := 2;
  CheckLength;
  FlclPoints[0].X := 0;
  FlclPoints[0].Y := 0;
  FlclPoints[1].X := 0;
  FlclPoints[1].Y := 10;
end;

procedure TRakLines.Cut(Idx: integer);
begin
  NewGroupStart(TRakCommandGroup.Create);
  try
    if (0 <= Idx) and (Idx < Count-1) then
    begin
      CheckLength(1);

      InsertPoints(Idx+1, 1);

      Points[Idx+1] := ExtPoint((Points[Idx].X+Points[Idx+2].X)/2, (Points[Idx].Y+Points[Idx+2].Y)/2);
    end;
  finally
    NewGroupStop;
  end;
end;

procedure TRakLines.DeletePoint(const Idx: integer);
begin
  if IsBetween(Idx, 0, Count-1) then
  begin
    if Count < 3 then
    begin // Wenn wir hchstens eine Linie haben
      // Dann wird alles gelscht
      Clear;
    end
    else
    begin
      DeletePoints(Idx, 1);
    end;
  end;
end;

procedure TRakLines.DeletePoints(Idx, Anz: integer);
var
  i: integer;
begin
  if Idx < Count then
  begin
    if Idx + Anz > Count then
      Anz := Count - Idx;
    for i := Idx to Count - Anz - 1 do
      Points[i] := Points[i+Anz];

    Count := Count - Anz;
  end;
end;

procedure TRakLines.GetBounds(out Links, Oben, Breite, Hoehe: extended);
var
  X, Y, Rechts, Unten: extended;
  i: integer;
begin
  X := Points[0].X;
  Y := Points[0].Y;
  Links := X;
  Oben := Y;
  Rechts := X;
  Unten := Y;
  for i := 1 to PointsCount - 1 do
  begin
    X := Points[i].X;
    Y := Points[i].Y;

    if X < Links then
      Links := X
    else if X > Rechts then
      Rechts := X;

    if Y < Oben then
      Oben := Y
    else if Y > Unten then
      Unten := Y;
  end;

  Breite := Rechts - Links;
  Hoehe := Unten - Oben;
end;

function TRakLines.GetPointIdxAt(X, Y: extended): integer;
var
  i: integer;
begin
  Result := -1;
  for i := 0 to PointsCount - 1 do
  begin
    if (Points[i].X = X) and (Points[i].Y = Y) then
    begin
      Result := i;
      break;
    end;
  end;
end;

function TRakLines.GetPoints(Idx: integer): TExtPoint;
begin
  Result := FlclPoints[Idx];
end;

function TRakLines.GetPointsCount: integer;
begin
  Result := Count;
end;

procedure TRakLines.InsertPoint(const Idx: integer;
  const NewPoint: TExtPoint);
begin
  NewGroupStart(TRakCommandGroup.Create);
  try
  	InsertPoints(Idx, 1);
    Points[Idx] := NewPoint;
  finally
    NewGroupStop;
  end;
end;

procedure TRakLines.InsertPoints(Idx, Anz: integer);
var
  i: integer;
begin
  if Idx <= Count then
  begin
    CheckLength(Anz);
    for i := Count + Anz - 1 downto Idx + Anz do
      Points[i] := Points[i-Anz];

    Count := Count + Anz;
  end;
end;

procedure TRakLines.LoadFromBinaryTag(Tag: TRakBinaryTag);
var
  i: integer;
  List: IRakBinaryTagList;
  SubTag: TRakBinaryTag;
begin
  if Tag <> nil then
  begin
    List := Tag.FindTagsByName('Points');
    Count := List.Count;
    CheckLength(10);
    for i := 0 to Count - 1 do
    begin
      SubTag := List.Items(i);
      Points[i] := ExtPoint(SubTag.AttributeAsFloat('X', 0),
        SubTag.AttributeAsFloat('Y', 0));
    end;
  end;
end;

procedure TRakLines.Mirror(Horizontal, Vertical: boolean; XCenter,
  YCenter: extended);
var
  i: integer;
  APoint: TExtPoint;
begin
  if Horizontal or Vertical then
  begin
    NewGroupStart(TRakCommandGroup.Create);
    try
      for i := 0 to PointsCount - 1 do
      begin
        APoint := Points[i];

        if Horizontal then
          APoint.X := 2*XCenter - APoint.X;

        if Vertical then
          APoint.Y := 2*YCenter - APoint.Y;

        Points[i] := APoint;
      end;
    finally
      NewGroupStop;
    end;
  end;
end;

procedure TRakLines.Move(OffsetX, OffsetY: extended);
var
  i: integer;
begin
  NewGroupStart(TRakCommandGroup.Create);
  try
    for i := 0 to PointsCount - 1 do
      Points[i] := ExtPoint(Points[i].X + OffsetX, Points[i].Y + OffsetY);
  finally
    NewGroupStop;
  end;
end;

procedure TRakLines.MoveToRect(DestRect: TExtRect);
var
  i: integer;
  Links, Oben, Breite, Hoehe: extended;
  XScale, YScale: extended;
  APoint: TExtPoint;
begin
  NewGroupStart(TRakCommandGroup.Create);
  try
    GetBounds(Links, Oben, Breite, Hoehe);
    DestRect := NormalizeRect(DestRect);

    if Breite = 0 then
      XScale := 0
    else
      XScale := (DestRect.Right - DestRect.Left) / Breite;

    if Hoehe = 0 then
      YScale := 0
    else
      YScale := (DestRect.Bottom - DestRect.Top) / Hoehe;

    for i := 0 to PointsCount - 1 do
    begin
      APoint := Points[i];
      APoint.X := (APoint.X - Links) * XScale + DestRect.Left;
      APoint.Y := (APoint.Y - Oben) * YScale + DestRect.Top;
      Points[i] := APoint;
    end;
  finally
    NewGroupStop;
  end;
end;

procedure TRakLines.Pack(von, bis: integer);
var
  i: integer;
begin
  NewGroupStart(TRakCommandGroup.Create);
  try
    if Count > 2 then
    begin
      KleinerGroesser(von, bis);

      i := bis;
      while (i > von) and (Count > 2) do
      begin
        if ExtPointsEqual(Points[i], Points[i-1]) then
          DeletePoints(i, 1);

        dec(i);
      end;
    end;
  finally
    NewGroupStop;
  end;
end;

procedure TRakLines.SaveToBinaryTag(Tag: TRakBinaryTag);
var
  i: integer;
  SubTag: TRakBinaryTag;
begin
  for i := 0 to Count - 1 do
  begin
    SubTag := Tag.AddTag('Points');
    TRakBinaryExtended(SubTag.AddAttribute(TRakBinaryExtended, 'X')).Value := Points[i].X;
    TRakBinaryExtended(SubTag.AddAttribute(TRakBinaryExtended, 'Y')).Value := Points[i].Y;
  end;
end;

procedure TRakLines.SetCount(const Value: integer);
var
  Alt: integer;
begin
  if FlclCount <> Value then
  begin
    Alt := FlclCount;
    FlclCount := Value;
    AddCommand(TRakHvlHcFCount.Init(Alt, Value));
  end;
end;

end.
