unit RakTreeViewBox;

{///////////////////////////////////////////////////////////////////////////////
// Beschreibung : Die Unit RakTreeViewBox stellt Klassen fr die
                  die Kapselung eines TTreeView-Objektes zur Verfgung.
                  Bei TreeViews hat man oft das Problem, dass der angezeigte
                  Text eines Knotens kein eindeutiger Schlssel fr diesen Knoten ist.
                  Man kann zwar an jeden Knoten eine Referenz eines Objektes
                  anhngen (TTreeNode.Data), aber der Knoten wird
                  nicht Eigentmer des Objektes und gibt es somit auch nicht
                  wieder frei.
                  Hier werden Klassen zur Verfgung gestellt, die eine TTreeView
                  kapseln, fr jeden Knoten ein Objekt fr weitere Daten zur
                  Verfgung stellen und sich auch um die Freigabe des
                  Objektes kmmert.
                  Ausserdem knnen einzelne Knoten unsichtbar gemacht
                  werden, ohne dass die Verbindung zu den Daten verloren geht.

                  Hinweis: Wenn man auf seinem Formular einen neuen Baum
                          verwenden will, dann sollte man evtl. die
                          Komponente TRakTreeView (Unit RakTreeiew) anstelle
                          von TTreeView/TRakTreeViewBox
                          verwenden
// erstellt am  : ???
////////////////////////////////////////////////////////////////////////////////
// nderungen:
///////////////////////////////////////////////////////////////////////////////}
// Delphi 5 Implementation
//
// Copyright (c) 2000 by Richard Kaspar
// ("http://www.kasparsoft.de")
//
// Definitions:
// - "Package" refers to the collection of files distributed by
//   the Copyright Holder, and derivatives of that collection of
//   files created through textual modification.
// - "Standard Version" refers to such a Package if it has not
//   been modified, or has been modified in accordance with the
//   wishes of the Copyright Holder.
// - "Copyright Holder" is whoever is name in the copyright or
//   copyrights for the package.
// - "You" is you, if you're thinking about copying or distributing
//   this Package.
//
// Permission is hereby granted, free of charge, to any person
// obtaining a copy of this software and associated documentation
// files (the "Package"), to deal in the Package without restriction,
// including without limitation the rights to use, copy, modify,
// merge, publish, distribute, sublicense, and/or sell copies of the
// Package, and to permit persons to whom the Package is furnished
// to do so, subject to the following conditions:
//
// The above copyright notice and this permission notice shall be
// included in all copies or substantial portions of the Package.
//
// You may modify your copy of this Package in any way, provided
// that you insert a prominent notice in each changed file stating
// how and when you changed a file, and provided that you do at
// least one of the following:
//
// a) allow the Copyright Holder to include your modifications in
// the Standard Version of the Package.
//
// b) use the modified Package only within your corporation or
// organization.
//
// c) rename any non standard executables, units, and classes so
// the names do not conflict with standard executables, units, and
// classes, and provide a separate manual page that clearly documents
// how it differs from the standard version.
//
// d) make other distribution arrangements with the Copyright Holder.
//
// The name of the Copyright Holder may not be used to endorse or
// promote products derived from this Package without specific prior
// written permission.
//
// THE PACKAGE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
// EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
// MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.
// IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY
// CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT,
// TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE
// PACKAGE OR THE USE OR OTHER DEALINGS IN THE PACKAGE.
{$I RAK.INC}

interface

uses Forms, Classes, ComCtrls, RakCollection, SysUtils;

type
{////////////////////////////////////////////////////////////////////////////////
// Klasse       : TRakNode
// Beschreibung : TRakNode kapselt einen Knoten eines TTreeViews und damit
                  den auch den gesammten Baum.
                  Am Anfang erstellt man eine Instanz von TRakNode. Dem Property
                  Anker weisst man TTreeView.Items zu.
                  Jetzt knnen mit der Methode Add dem Baum weiter Knoten
                  hinzugefgt werden. Ein Knoten wird einfach durch den Aufruf
                  von seiner Free-Methode aus dem Baum gelscht.
// erstellt am  :
////////////////////////////////////////////////////////////////////////////////
// nderungen:
////////////////////////////////////////////////////////////////////////////////}
  TRakNode = class(TRakCollection)
  private
    FAnker: TTreeNodes;
    FNode: TTreeNode;
    FSearchedTvwNode: TTreeNode;
    FText: string;
    FVisible: boolean;

    FId: integer;
    FStringId: string;

    function GetTvwAnker: TTreeNodes;
    procedure SetTvwAnker(Value: TTreeNodes);
    function GetItems(Index: Integer): TRakNode;
    function GetNode: TTreeNode;
    function GetParentNode: TRakNode;
    function GetText: string;
    procedure SetText(Value: string);
    procedure SetVisible(Value: boolean);
  protected
    function CompareItems(Item1, Item2: Pointer): integer; override;
    procedure DoDestroy; override;
    function IsSearched(Item: TRakCollection): boolean; override;

    procedure DeleteNodes; virtual;
    procedure MoveNode; virtual;
    procedure NodeDeletion(Sender: TObject; Node: TTreeNode); virtual;

    function GetBaseNode: TRakNode; virtual;
    procedure SetCollection(Value: TRakCollection); override;
    procedure SetIndex(Value: integer); override;
    function GetSelected: TRakNode; virtual;
    procedure SetSelected(Value: TRakNode); virtual;
  public
    constructor Create; override;

    // Add fgt dem Knoten einen neuen Unterknoten hinzu
    function Add: TRakNode;
    // Assign kopiert den ganzen Zweig
    procedure Assign(Source: TRakCollectionItem); override;
    // ItemByText sucht einen direkten Unterknoten anhand seines angezeigten Textes
    function ItemByText(Text: string): TRakNode;
    // ItemByTvwNode durchsucht den ganzen Baum nach dem Knoten, der einen bestimmten Treeviewknoten reprsentiert
    function ItemByTvwNode(Node: TTreeNode): TRakNode;
    // MoveTvwNodes verschiebt einen ganzen Zweig
    procedure MoveTvwNodes(FromNode, ToNode: TTreeNode; Mode: TRakItemMoveMode);
    // Sucht einen direkten Unterknoten Anhand seines Propertis "Id".
    function NodeById(Idx: integer): TRakNode;
    // Sucht einen direkten Unterknoten Anhand seines Propertis "StringId".
    function NodeByStringId(sId: string): TRakNode;
    // Rebuild sorgt dafr, dass der kapselnde Zweig (TRakNode) wieder dem gekapselten Zweig (TTreeNode) entspricht, falls direkt ber Methoden von TTreeNode etwas verndert wurde
    procedure Rebuild; virtual;
    // WiederAufsetzen ist wie Rebuild, nur dass es schneller ist, aber nur in bestimmten Situationen verwendet werden kannn
    procedure WiederAufsetzen; virtual;
    // Sort sortiert die Unterknoten alphabetisch
    procedure Sort; override;

    // Der Basisknoten des gekapselten Baumes
    property Anker: TTreeNodes read GetTvwAnker write SetTvwAnker;
    // Der Basisknoten des kapselnden Baumes
    property BaseNode: TRakNode read GetBaseNode;
    // Die Unterknoten (0 <= Index < Count)
    property Items[Index: Integer]: TRakNode read GetItems; default;
    // Der gekapselte Knoten
    property Node: TTreeNode read GetNode;
    // Der berknoten
    property ParentNode: TRakNode read GetParentNode;
    // Der selektierte Knoten
    property Selected: TRakNode read GetSelected write SetSelected;
    // Der angezeigte Text
    property Text: string read GetText write SetText;
    // Legt fest, ob der Knoten angezeigt wird
    property Visible: boolean read FVisible write SetVisible;

    // Frei vergebare Information
    property Id: integer read FId write FId;
    property StringId: string read FStringId write FStringId;
  end;

implementation

var
  Dummy: integer;

constructor TRakNode.Create;
begin
  inherited Create;

  FVisible := true;
end;

function TRakNode.CompareItems(Item1, Item2: Pointer): integer;
  function SortString(Str: string): string;
  var
    i: integer;
  begin
    Result := '';
    for i := 1 to length(Str) do
    begin
      case Str[i] of
        '','': Result := Result + 'AE';
        '','': Result := Result + 'OE';
        '','': Result := Result + 'UE';
        '': Result := Result + 'SS';
        else Result := Result + Str[i];
      end;
    end;
  end;
var
  Child1, Child2: TRakNode;
begin
  Child1 := TRakNode(Item1);
  Child2 := TRakNode(Item2);
  Result := CompareText(SortString(Child1.Text), SortString(Child2.Text));
end;

procedure TRakNode.DoDestroy;
begin
  inherited DoDestroy;

  try
    if Collection = nil then
      Anker := nil
    else
    begin
      if (Node <> nil) and (Anker <> nil) and
        not(Classes.csDestroying in Anker.Owner.ComponentState) and
        not (Classes.csDestroying in Application.ComponentState) then
      begin
        Text := Node.Text;
        // Die nchste Zeile macht eigentlich nichts, auer da sie
        // verhindert, da das Programm in der bernchsten Zeile
        // hngen bleibt.
        Dummy := Node.Index;
        Node.Free;
        FNode := nil;
      end;
    end;
  except;
  end;
end;

function TRakNode.Add: TRakNode;
begin
  Result := TRakNode(inherited Add(TRakCollectionItemClass(ClassType)));
  Result.FNode := Anker.AddChild(Node, 'NewNode');
end;

procedure TRakNode.Assign(Source: TRakCollectionItem);
begin
  inherited Assign(Source);
  FNode := TRakNode(Source).FNode;
end;

function TRakNode.ItemByText(Text: string): TRakNode;
var
  i: integer;
begin
  Result := nil;
  i := 0;
  while (Result = nil) and (i < Count) do
  begin
    if Items[i].Text = Text then
      Result := Items[i];
    inc(i);
  end;
end;

function TRakNode.NodeByStringId(sId: string): TRakNode;
var
  i: integer;
begin
  if StringId = sId then
    Result := self
  else
  begin
     Result := nil;
    i := 0;
    while (Result = nil) and (i < Count) do
    begin
      Result := Items[i].NodeByStringId(sId);
      inc(i);
    end;
  end;
end;

function TRakNode.NodeById(Idx: integer): TRakNode;
var
  i: integer;
begin
  if Id = Idx then
    Result := self
  else
  begin
     Result := nil;
    i := 0;
    while (Result = nil) and (i < Count) do
    begin
      Result := Items[i].NodeById(Idx);
      inc(i);
    end;
  end;
end;

function TRakNode.ItemByTvwNode(Node: TTreeNode): TRakNode;
begin
  if Node = nil then
    Result := nil
  else
  begin
    FSearchedTvwNode := Node;
    Result := TRakNode(FindTreeItem);
  end;
end;

function TRakNode.IsSearched(Item: TRakCollection): boolean;
begin
  Result := FSearchedTvwNode = TRakNode(Item).Node;
end;

function TRakNode.GetParentNode: TRakNode;
begin
   Result := TRakNode(Collection);
end;

procedure TRakNode.DeleteNodes;
var
  i: integer;
begin
  FNode := nil;
  for i := 0 to Count -1 do
    Items[i].DeleteNodes;
end;

procedure TRakNode.NodeDeletion(Sender: TObject; Node: TTreeNode);
var
  ANode: TRakNode;
begin
  if Classes.csDestroying in FAnker.Owner.ComponentState then
    Anker := nil
  else
  begin
    ANode := ItemByTvwNode(Node);
    if ANode <> nil then
      ANode.FNode := nil;
  end;
end;

function TRakNode.GetBaseNode: TRakNode;
begin
  if Collection = nil then
    Result := self
  else
    Result := TRakNode(Collection).GetBaseNode;
end;

procedure TRakNode.SetSelected(Value: TRakNode);
begin
  if Assigned(Value.FNode) then
    Value.FNode.Selected := true;
end;

function TRakNode.GetSelected: TRakNode;
var
  i: integer;
begin
  Result := nil;
  if not (csDestroying in State) and Visible then
  begin
    if Assigned(FNode) and Node.Selected then
      Result := self
    else
    begin
      i := 0;
      while (Result = nil) and (i < Count) do
      begin
        Result := Items[i].Selected;
        inc(i);
      end;
    end;
  end;
end;

procedure TRakNode.SetCollection(Value: TRakCollection);
begin
  inherited SetCollection(Value);

  if (Node <> nil) and (Collection <> nil) then
    Node.MoveTo(TRakNode(Collection).Node, naAddChild);
end;

procedure TRakNode.MoveNode;
begin
  if Node <> nil then
  begin
    if Index = Collection.Count - 1 then
      Node.MoveTo(TRakNode(Collection).Node, naAddChild)
    else
    begin
      Node.MoveTo(TRakNode(Collection.Items[Index + 1]).Node, naInsert);
    end;
  end;
end;

procedure TRakNode.SetIndex(Value: integer);
begin
  inherited SetIndex(Value);

  MoveNode;
end;

procedure TRakNode.MoveTvwNodes(FromNode, ToNode: TTreeNode;
  Mode: TRakItemMoveMode);
begin
  ItemByTvwNode(FromNode).MoveTo(ItemByTvwNode(ToNode), Mode)
end;

procedure TRakNode.WiederAufsetzen;
  procedure NeuAufsatz(ANode: TRakNode; var i: integer);
  var
    t: integer;
  begin
    t := 0;
    while t < ANode.Count do
    begin
      ANode[t].FNode := FAnker.Item[i];
      inc(i);

      NeuAufsatz(ANode[t], i);

      inc(t);
    end;
  end;
var
  i: integer;
begin
  if GetBaseNode = self then
  begin
    i := 0;
    NeuAufsatz(self, i);
  end;
end;

procedure TRakNode.Rebuild;
var
  i: integer;
  ANode: TRakNode;
begin
  if ParentNode = nil then
  begin
    if Assigned(FAnker) then
    begin
      DeleteNodes;
      FreeAll;
      for i := 0 to FAnker.Count - 1 do
      begin
        if FAnker[i].Parent = nil then
        begin
          ANode := TRakNode(inherited Add(TRakCollectionItemClass(ClassType)));
          ANode.FNode := FAnker[i];
          ANode.Rebuild;
        end;
      end;
    end;
  end
  else
  begin
    for i := 0 to Node.Count - 1 do
    begin
      ANode := TRakNode(inherited Add(TRakCollectionItemClass(ClassType)));
      ANode.FNode := Node[i];
      ANode.Rebuild;
    end;
  end;
end;

procedure TRakNode.Sort;
var
  i, t: integer;
  List: TList;
begin
  List := TList.Create;
  try
    for i := 0 to Count -1 do
      List.Add(Items[i].Node);

    inherited Sort;

    for i := 0 to Count - 2 do
    begin
      if (Items[i].Node <> nil) and (Items[i].Node <> List[i]) then
      begin
        for t := i + 1 to Count - 1 do
        begin
          if List[t] <> nil then
          begin
            Items[i].Node.MoveTo(TTreeNode(List[t]), naInsert);
            List.Move(List.IndexOf(Items[i].Node), t);
            break;
          end;
        end;
      end;
    end;
  finally
    List.Free;
  end;
end;

function TRakNode.GetTvwAnker: TTreeNodes;
begin
  if Collection = nil then
    Result := FAnker
  else
    Result := TRakNode(GetAnker).Anker;
end;

procedure TRakNode.SetTvwAnker(Value: TTreeNodes);
begin
  if Collection = nil then
  begin
    if Assigned(FAnker) then
    begin
      TTreeView(FAnker.Owner).OnDeletion := nil;
      DeleteNodes;
    end;

    FAnker := Value;

    if Assigned(FAnker) then
    begin
      TTreeView(FAnker.Owner).OnDeletion := NodeDeletion;
      Rebuild;
    end;
  end;
end;

function TRakNode.GetItems(Index: Integer): TRakNode;
begin
  Result := TRakNode(inherited Items[Index]);
end;

function TRakNode.GetNode: TTreeNode;
begin
  Result := FNode;
end;

function TRakNode.GetText: string;
begin
  if Node <> nil then
    Result := Node.Text
  else
    Result := FText;
end;

procedure TRakNode.SetVisible(Value: boolean);
var
  i: integer;
begin
  if Value <> Visible then
  begin
    FVisible := Value;

    if Visible then
    begin
      if Collection <> nil then
        FNode := Anker.AddChild(TRakNode(Collection).Node, FText);
    end;

    for i := 0 to Count - 1 do
      Items[i].Visible := Visible;

    if not Visible then
    begin
      FText := Text;
      FNode.Free;
      FNode := nil;
    end;
  end;
end;

procedure TRakNode.SetText(Value: string);
begin
  FText := Value;

  Node.Text := Value;
end;


end.
